/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


// TO DO LIST
// - Add support for FreeType2.

// #define DEBUG_EVENT_LOOP

extern "C" {
#include <math.h>
#include <stdio.h>
#include <stdlib.h>	// exit, getenv
#include <string.h>
#include <ctype.h>	// isspace
#include <unistd.h>	// usleep
#include <sys/time.h>	// gettimeofday
#include <TouchWidgets/Util.h>
}

#include <TouchWidgets/Common.h>
#include <TouchWidgets/App.h>
#include <TouchWidgets/Container.h>

using namespace TouchWidgets;

#define default_icon_width 16
#define default_icon_height 16
#define default_icon_x_hot 7
#define default_icon_y_hot 7
static char default_icon_bits[] = {
   0x00, 0x00, 
   0x00, 0x00, 
   0x00, 0x00, 
   0x00, 0x00, 
   0x00, 0x00, 
   0xc0, 0x01,
   0x60, 0x03, 
   0x20, 0x02, 
   0x60, 0x03, 
   0xc0, 0x01, 
   0x00, 0x00, 
   0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

// #define DEGREE_SIGN_CHAR 176

App *app;
static int run_before_main (void)
{
	app = NULL;
	return 0;
}
static int dummyValue = run_before_main ();

//--------------------

void
App::present_root_controller (Controller *controller)
{
	controller->will_appear ();
	Widget *rw = controller->get_root_widget ();
	if (rw)
		rw->set_rect (Rect_new (0, 0, controller->size.width, controller->size.height));
	controller->layout_widgets ();
	Window window = get_top_window ();
	controller->create_windows ();
	controller->did_appear ();
}

int
App::process_event (XEvent *e)
{
	if (n_windows < 1)
		return 0;
	if (!e) 
		return 0;
	//==========

	XAnyEvent *any = (XAnyEvent*) e;
	XExposeEvent *expose = (XExposeEvent*) e;
	XKeyEvent *key = (XKeyEvent*) e;

	Display *affected_dpy = any->display;
	if (affected_dpy != dpy)
		return 0;

	// RULE: If a popup window exists, don't process events (except Expose)
	// to the main window.

	Window top_window = windows [n_windows-1];
	Window affected_window = any->window;
	if (affected_window != top_window && n_windows > 1) {
		if (e->type != Expose && e->type != KeyPress && e->type != KeyRelease) {
#if 0
			if (e->type == FocusIn && top_window != affected_window) {
				XLowerWindow (dpy, affected_window);
				XRaiseWindow (dpy, top_window);
				XFlush (dpy);
			}
#endif
			return 0;
		}
	}

	// RULE: If the user has tapped on a window that is not
	// on the top, causing it to have focus, raise the top-of-stack window.
	//

	// RULE: If top window doesn't have a controller attached yet,
	// we have to ignore any events for the window.

	Controller *top_controller = controllers [n_windows-1];
	if (!top_controller)
		return 0;

	// Determine if the affected window is an uninitialized top level window.
	for (int i = 0; i < n_windows; i++) {
		if (affected_window == windows[i] && !window_is_initialized [i]) {
			Controller *c = get_root_controller_for_window (affected_window);
			if (c) {
				present_root_controller (c);
				window_is_initialized [i] = true;
			}
		}
	}

	switch (e->type) {
	case DestroyNotify:
#ifdef DEBUG_EVENT_LOOP
		puts ("Event: DestroyNotify");
#endif
		return -1;
	
	case Expose: {
		int x0,y0,x1,y1;
		x0 = expose->x;
		y0 = expose->y;
		x1 = x0 + expose->width - 1;
		y1 = y0 + expose->height - 1;
#ifdef DEBUG_EVENT_LOOP
		printf ("Event: Expose of window %08lx, rect = %dx%d @ %d,%d\n", (unsigned long) affected_window,expose->width,expose->height,x0,y0);
#endif

		Rect r = Rect_new (x0, y0, expose->width, expose->height);
		int i;
		for (i = 0; i < n_windows; i++) {
			Controller *c = controllers[i];
			Window win = windows [i];
			Widget *w;
			if (win == affected_window) {
				w = c->get_root_widget ();
				if (w) {
					w->draw (r);
					break;
				}
			} else {
				Widget *w = c->get_widget_for_window (affected_window);
				if (w) {
					w->draw (r);
					break;
				}
			}
		}
		if (i == n_windows) {
			// app->log_warning ("Cannot find widget for window being exposed.");
		}

		alt_pressed = meta_pressed = false;
		} break;
	
	case ResizeRequest: {
		XResizeRequestEvent *resize = (XResizeRequestEvent*) e;
#ifdef DEBUG_EVENT_LOOP
		printf ("Event: ResizeRequest of window %08lx: %dx%d\n", (unsigned long) affected_window, resize->width,resize->height);
#endif
		requested_window_width = resize->width;
		requested_window_height = resize->height;

		top_controller->set_rect (Rect_new (0, 0, requested_window_width, requested_window_height));
		top_controller->layout_widgets ();
		} break;

	case UnmapNotify: {
#ifdef DEBUG_EVENT_LOOP
		printf ("Event: UnmapNotify of window 0x%lx\n", (unsigned long) affected_window);
#endif
		} break;

	case MapNotify: {
#ifdef DEBUG_EVENT_LOOP
		printf ("Event: MapNotify of window 0x%lx\n", (unsigned long) affected_window);
#endif
		} break;

	case KeyRelease: {
#ifdef DEBUG_EVENT_LOOP
		puts ("Event: KeyRelease");
#endif
		KeySym sym = XKeycodeToKeysym (dpy, key->keycode, 0);
		switch (sym) {
		case XK_Alt_L:
		case XK_Alt_R:
			// NOTE: On OS/X in XQuartz, don't forget to change default
			// setting for Option keys to send Alt keycodes.
			puts ("ALT key is UP.");
			alt_pressed = false;
			break;
		case XK_Meta_L: 
		case XK_Meta_R: 
			// NOTE: On OS/X this is the Command key.
			puts ("META key is UP.");
			meta_pressed = false;
			break;
		default:
			// RULE: As regards keys from the physical keyboard (e.g. from a USB keyboard),
			// if the on-screen keyboard exists, it will receive keystrokes
			// rather than the key-focus widget.

			if (keyboard) {
				send_message (keyboard, KEY_UP, sym);
			}
			else if (key_focus_widget) {
				send_message (key_focus_widget, KEY_UP, sym);
			}
		}
		} break;

	case KeyPress: {
#ifdef DEBUG_EVENT_LOOP
		puts ("Event: KeyPress");
#endif
		KeySym sym = XKeycodeToKeysym (dpy, key->keycode, 0);
		switch (sym) {
		case XK_Alt_L:
		case XK_Alt_R:
			// NOTE: On OS/X in XQuartz, don't forget to change default
			// setting for Option keys to send Alt keycodes.
			puts ("ALT key is DOWN.");
			alt_pressed = true;
			break;
		case XK_Meta_L: 
		case XK_Meta_R: 
			// NOTE: On OS/X this is the Command key.
			puts ("META key is DOWN.");
			meta_pressed = true;
			break;
#if 0
				if (sym == w->accelerator)
				{
					if (w->user_callback)
					{
						(w->user_callback) (w,
							KEY_DOWN);
						break;
					}
				}
				w = w->next;
			}
#endif
		default:
			if (keyboard) {
				send_message (keyboard, KEY_DOWN, sym);
			}
			else
			if (key_focus_widget) {
				send_message (key_focus_widget, KEY_DOWN, sym);
			}
		}
		} break;
	
	case EnterNotify: {
#ifdef DEBUG_EVENT_LOOP
		XEnterWindowEvent *enter = (XEnterWindowEvent*) e;
		printf ("Event: EnterNotify for window %d\n", (unsigned long) affected_window);
#endif
		} break;
	
	case LeaveNotify: {
#ifdef DEBUG_EVENT_LOOP
		XLeaveWindowEvent *leave = (XLeaveWindowEvent*) e;
		printf ("Event: LeaveNotify for window %d\n", (unsigned long) affected_window);
#endif
		} break;
	
	case MotionNotify: {
		XMotionEvent *motion = (XMotionEvent*) e;
		int x = motion->x;
		int y = motion->y;
		Time t = motion->time;
		if (touch_focus_widget) {
			// Make (x,y) relative to the widget that was originally touched.
			//
			Point p = Widget::convert_absolute_point_to_relative (Point_new (x,y), touch_focus_widget);
			update_touch_speed_estimation (t, p);
#ifdef DEBUG_MOMENTUM
			printf ("Estimated touch speed %g pixels/ms\n", get_touch_speed_estimation ());
#endif
			touch_focus_widget->touches_continue (&p, 1);
#ifdef DEBUG_EVENT_LOOP
		printf ("Event: MotionNotify at %d,%d\n", x, y);
#endif
		} 
		} break;

	case ButtonPress: {
//		if (root_controller) {
			XButtonEvent *button = (XButtonEvent*) e;
			int x = button->x;
			int y = button->y;
			Time t = button->time;
			Point p = Point_new (x, y);
#ifdef DEBUG_EVENT_LOOP
			printf ("Touch down at %d,%d\n", x, y);
#endif
			Widget *w = top_controller->get_root_widget ();
			if (!w) {
				log_warning ("Controller.has no root widget.");
			} else {
				Point p2 = p;
				w = app->identify_widget_at_absolute_point (p, &p2);
				if (!w) {
					log ("No widget touched.");
				}
#ifdef DEBUG_EVENT_LOOP
				printf ("Widget touched is %s.\n", w->name);
#endif
				touch_focus_widget = w;
				send_message (touch_focus_widget, TOUCH_DOWN_INSIDE, MessageParameter_from_point (p2));

				begin_touch_speed_estimation (t, p2);
				w->touches_start (&p2, 1);
			}
//		}
		} break;

	case ButtonRelease: 
		if (top_controller && touch_focus_widget) {
			XButtonEvent *button = (XButtonEvent*) e;
			int x = button->x;
			int y = button->y;
			Point p = Point_new (x, y);

#ifdef DEBUG_EVENT_LOOP
			printf ("Touch up at %d,%d\n", x, y);
#endif
			Rect r = touch_focus_widget->get_rect ();
			Point relative_coords = Widget::convert_absolute_point_to_relative (Point_new (x,y), touch_focus_widget);
			touch_focus_widget->touches_end (&relative_coords, 1);
			if (relative_coords.x >= 0 && relative_coords.y >= 0 && relative_coords.x < r.size.width 
				&& relative_coords.y < r.size.height)
			{
				send_message (touch_focus_widget, TOUCH_UP_INSIDE, MessageParameter_from_point (relative_coords));
			} else {
				send_message (touch_focus_widget, TOUCH_UP_OUTSIDE);
			}
			touch_focus_widget = NULL;
		}
		break;
	
	case ReparentNotify:	/* puts ("ReparentNotify"); */ break;
	case ConfigureNotify:	
#ifdef DEBUG_EVENT_LOOP
		puts ("ConfigureNotify"); 
#endif
		break;
	case CirculateNotify:	/* puts ("CirculateNotify"); */ break;
	case GravityNotify:	/* puts ("GravityNotify"); */ break;

	default:
#ifdef DEBUG_EVENT_LOOP
		printf ("Other event: %d\n", e->type);
#endif
		return -2;
	}

	return 0;
}

//-----------------------------------------------------------------------------
// Name:	run
// Purpose:	This is the event loop.
//-----------------------------------------------------------------------------

void App::run ()
{
	XEvent ev;

	while (!quitnow) {
		uint32_t t0 = get_microsecond_time ();

		//--------------------
		// Process all events.
		//
		while (XCheckMaskEvent (app->dpy, 0xffffffff, &ev)) {
			int retval = process_event (&ev);
			if (retval == -1) {
				log ("Got destroy event.");
				fflush (NULL);
				break;
			}
			if (retval == -2)
				log_warning ("Event ignored");
		}

		//--------------------
		// Continue animations.
		//
		if (n_animations) {
			for (unsigned i = 0; i < n_animations; i++) {
				Widget *w = animations [i];
				if (w->is_animating ()) {
#ifdef DEBUG_ANIMATION
					w->print_animation_diagnostics ();
#endif
					w->generate_next_animation_rect ();
				} else {
					// Remove it.
					unsigned ix = i;
					while (ix < n_animations-1) {
						animations [ix] = animations [ix+1];
						ix++;
					}
					--n_animations;
				}
			}
		}

		//--------------------
		// Internal animations.
		//
		// RULE: Since the momentum effect is tied to touches, which are
		// only directed at one Widget at this time, .'. only one
		// widget is allowed to do internal animation at a time.
		//
		if (internally_animating_widget) {
			internally_animating_widget->perform_internal_animation ();
		}

		//-------------------------
		// Convey indirect messages.
		//
		if (n_indirect_messages) {
			int read_index = 0;
			int write_index = 0;
			uint32_t now = get_millisecond_time ();
			while (read_index < n_indirect_messages) {
				MessageBundle *m = indirect_messages [read_index++];
				if (m->fire_time > now) {
					indirect_messages [write_index++] = m; // Save for later.
				} else {
// XX Timer goes here.
					Object *obj = m->recipient;
					if (obj && m->message) {
						obj->receive_message (m->sender, m->message, m->param);
					}
					delete m;
				}
			}
			n_indirect_messages = write_index;
		}

		//--------------------
		// Process new Exposes.
		//
		XSynchronize (dpy, true);
		while (XCheckMaskEvent (app->dpy, ExposureMask, &ev)) {
			process_event (&ev);
		}

		//--------------------
		// Sleep.
		//
		XFlush (dpy);
		uint32_t t = get_microsecond_time ();
		if (t - t0 > MICROSECONDS_PER_EVENT_LOOP_ITERATION) {
			char tmp [200];
			sprintf (tmp, "Event processing took %g ms (limit is %g ms).", (t-t0)/1000.,
				MICROSECONDS_PER_EVENT_LOOP_ITERATION/1000.);
			app->log_warning (tmp);
		}
		else {
			uint32_t time_used = t - t0;
			usleep (MICROSECONDS_PER_EVENT_LOOP_ITERATION - time_used); 
		}
	}	
}

void
App::request_key_focus (Widget *w)
{
	if (!w) 
		return;

	key_focus_widget = w;

	keyboard = new Keyboard ();
	keyboard->name = "keyboardController";
	keyboard->set_delegate (this);
	keyboard->show ();
}

void
App::release_key_focus (Widget *w)
{
	if (!w) 
		return;

	if (key_focus_widget && key_focus_widget == w)
		key_focus_widget = NULL;

	// RULE: Releasing key focus does not dismiss the on-screen keyboard. That's what the dismiss_keyboard method is for.
}

void
App::convey_key_to_key_focus_widget (KeySym key, bool down)
{
	if (!key_focus_widget) {
		log_warning ("convey_key_to_key_focus_widget: No key focus widget!");
		return;
	}
	if (!keyboard) {
		log_warning ("convey_key_to_key_focus_widget: No on-screen keyboard!");
		return;
	}

	indirect_message (keyboard, key_focus_widget, down ? KEY_DOWN : KEY_UP, key);
}

Window 
App::get_root_window (Widget *rw)
{
	if (!rw)
		return None;
	if (n_windows < 1)
		return None;

	for (int i = 0; i < n_windows; i++) {
		Controller *c = controllers [i];
		if (rw == c->get_root_widget ())
			return windows [i];
	}
puts ("Unable to determine root window");
	return None;
}

Window 
App::get_window_for_root_controller (Controller *rc)
{
	if (!rc)
		return None;
	if (n_windows < 1)
		return None;
	//==========

	for (int i = 0; i < n_windows; i++) {
		Controller *c = controllers [i];
		if (rc == c)
			return windows [i];
	}
puts ("Unable to determine root window");
	return None;
}

Controller *
App::get_root_controller_for_window (Window w)
{
	if (!w)
		return NULL;
	if (n_windows < 1)
		return NULL;
	//==========

	for (int i = 0; i < n_windows; i++) {
		if (w == windows [i])
			return controllers [i];
	}
puts ("Unable to determine root window");
	return NULL;
}

void 
App::push_window (int type, Controller *c, int width, int height, RGB bgcolor, const char *window_title)
{
	if (n_windows >= MAX_WINDOWS) {
		log_warning ("Attempt made to create excessive windows.");
		return;
	}

	Window window = None;
	Window parent = XDefaultRootWindow (dpy);

	int x = 0, y = 0;
	switch (type) {
	case WINDOW_ALERT:
		x = (dpy_width - width)/2;
		y = (dpy_height - height)/2;
		break;
	case WINDOW_KEYBOARD:
		y = dpy_height - height;
		break;
	case WINDOW_MAIN:
		x = requested_window_x;
		y = requested_window_y;
		width = requested_window_width > 10 ? requested_window_width : dpy_width;
		height = requested_window_height > 10 ? requested_window_height : dpy_height;
		height -= 50; // Allow for window manager's windows.
		break;
	}

	controllers [n_windows] = c;
	window_types [n_windows] = type;
	c->size = Size_new (width, height);

	// RULE: Only the main window gets a border, and only if it is not root.
	int border_width = 0;
	int border_color = RGB_BLACK;
	if (type != WINDOW_ROOT && type != WINDOW_MAIN)
		border_width = 1;

	if (type == WINDOW_ROOT)
		window = parent;
	else
		window = XCreateSimpleWindow (dpy, parent, 
			x, y,
			width, 
			height, 
			border_width, 
			border_color, bgcolor);

        window_is_initialized [n_windows] = false;
	windows [n_windows] = window;
	n_windows++;

	// RULE: Only the main window gets an icon and is known to
	// the window manager.
	//
	if (type == WINDOW_MAIN) {
		Pixmap icon = XCreateBitmapFromData (dpy, window,
			default_icon_bits, default_icon_width, default_icon_height);

		XWMHints wm_hints;
		wm_hints.initial_state = NormalState;
		wm_hints.input = true;
		wm_hints.icon_pixmap = icon;
		wm_hints.flags = StateHint | IconPixmapHint | InputHint;

		XClassHint class_hints;
		class_hints.res_name = (char*) program_name;
		class_hints.res_class = (char*) program_name;

		XSizeHints size_hints;
		bzero ((void*) &size_hints, sizeof(XSizeHints));
		size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
		size_hints.width = width;
		size_hints.height = height;
		size_hints.min_width = width;
		size_hints.min_height = height;
		size_hints.max_width = width;
		size_hints.max_height = height;

		XTextProperty window_name;
		XTextProperty icon_name;
		const char *s1 = window_title ?: program_name;
		const char *s2 = window_title ?: program_name;
		XStringListToTextProperty ((char**) &s1, 1, (XTextProperty*) &window_name);
		XStringListToTextProperty ((char**) &s2, 1, (XTextProperty*) &icon_name);

		XSetWMProperties (dpy, window, &window_name, &icon_name,
				NULL, 0, &size_hints, &wm_hints,
				&class_hints);
	}
	// RULE: All other windows cannot be resized (for now).
	//
	else
	{
		XWMHints wm_hints;
		wm_hints.initial_state = NormalState;
		wm_hints.input = true;
		wm_hints.flags = StateHint | InputHint;

		XClassHint class_hints;
		class_hints.res_name = (char*) program_name;
		class_hints.res_class = (char*) program_name;

		XSizeHints size_hints;
		bzero ((void*) &size_hints, sizeof(XSizeHints));
		size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
		size_hints.width = width;
		size_hints.height = height;
		size_hints.min_width = width;
		size_hints.min_height = height;
		size_hints.max_width = width;
		size_hints.max_height = height;

		XTextProperty window_name;
		XTextProperty icon_name;
		const char *s1 = window_title ?: program_name;
		const char *s2 = window_title ?: program_name;
		XStringListToTextProperty ((char**) &s1, 1, (XTextProperty*) &window_name);
		XStringListToTextProperty ((char**) &s2, 1, (XTextProperty*) &icon_name);

		XSetWMProperties (dpy, window, &window_name, &icon_name,
				NULL, 0, &size_hints, &wm_hints,
				&class_hints);

		XSetWindowAttributes attr;
		bzero (&attr, sizeof(XSetWindowAttributes));
		attr.override_redirect = true;
		attr.save_under = true;
		XChangeWindowAttributes (app->dpy, window, CWOverrideRedirect | CWSaveUnder, &attr);
	}

	if (!colormap)
        	colormap = XCreateColormap (dpy, window, visual, 0);

	XSelectInput (dpy, window, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | PointerMotionMask | KeyReleaseMask | ResizeRedirectMask | StructureNotifyMask);

	if (type == WINDOW_ROOT)
		return;

	XMapRaised (dpy, window);

	//XMoveResizeWindow (dpy, window, x, y, width, height);
}

void 
App::pop_window ()
{
	if (n_windows < 1)
		return;

	n_windows--;
	Window window = windows [n_windows];
	Controller *controller = controllers [n_windows];
	delete controller;

	if (window_types [n_windows] == WINDOW_KEYBOARD) {
		keyboard = NULL;
	} 

	XDestroyWindow (dpy, window);
}


void
App::initialize (int argc, char **argv, 
		char *program_name,
		bool automatic,
		ushort window_w, 
		ushort window_h,
		uint32_t bgcolor,
		ushort border_size,
		uint32_t border_color)
{
	bool override_cursor_hiding = false;

	window_is_root = false;

	char *style_path = NULL;

	requested_window_x = 0;
	requested_window_y = 0;
	requested_window_width = 0;
	requested_window_height = 0;

	int i;
	for (i = 1; i < argc; i++) {
		char *s = argv [i];
		if (!strcmp ("-root", s))
			window_is_root = true;
		else if (!strcmp ("-cursor", s))
			override_cursor_hiding = true;
		else if (!strcmp ("-800", s)) {
			requested_window_width=800;
			requested_window_height=480;
			automatic = false;
		}
		else if (!strcmp ("-style", s)) {
			if (i != argc-1) {
				i++;
				style_path = argv [i];
			} else {
				app->log_fatal ("Missing style path.");
			}
		}
	}

#ifdef __APPLE__
    #include "TargetConditionals.h"
    #ifdef TARGET_OS_MAC
	override_cursor_hiding = true;
    #endif
#endif

	//------------------------------
	// Is style file openable?
	//
	FILE *f = fopen (style_path, "r");
	if (!f) {
		app->log_warning ("Cannot open specified style file.");
		style_path = NULL;
	}
	else {
		fclose (f);
	}

	//------------------------------
	// If no style file, use default.
	//
	if (!style_path) {
		snprintf (styles_path, PATH_MAX-1, "%s/.TouchWidgetsStyles", getenv ("HOME"));
	} else {
		strncpy (styles_path, style_path, PATH_MAX-1);
	}
	
	//-----------------------------------------------------
	// If the caller has specified excessively large window
	// dimensions, pare them down.
	//
	if (window_w > dpy_width)
		window_w = dpy_width - 2*border_size;
	if (window_h > dpy_height)
		window_h = dpy_height - 2*border_size;

	//-----------------------------
	// Get the default fonts.
	//
	default_title_font_bold = XLoadQueryFont (dpy, "-adobe-helvetica-bold-r-normal--20-140-100-100-p-105-iso8859-1");
	if (!default_title_font_bold) 
		default_title_font_bold = XLoadQueryFont (dpy, "10x20");
		//default_title_font_bold = XLoadQueryFont (dpy, "12x24");

	default_font = XLoadQueryFont (dpy, "-adobe-helvetica-medium-r-normal--17-120-100-100-p-88-iso8859-1");
	default_font_italic = XLoadQueryFont (dpy, "-adobe-helvetica-medium-o-normal--17-120-100-100-p-88-iso8859-1");
	default_font_bold_italic = XLoadQueryFont (dpy, "-adobe-helvetica-bold-o-normal--17-120-100-100-p-88-iso8859-1");
	default_font_bold = XLoadQueryFont (dpy, "-adobe-helvetica-bold-r-normal--17-120-100-100-p-88-iso8859-1");

	if (!default_font) {
		default_font = XLoadQueryFont (dpy, "9x18");
		if (!default_font)
			default_font = XLoadQueryFont (dpy, "8x13");
		if (!default_font)
			default_font = XLoadQueryFont (dpy, "fixed");
		if (!default_font) 
			log_fatal ("Cannot load basic font!");
	}

	if (!default_font) {
		default_font = XLoadQueryFont (dpy, "9x18b");
		if (!default_font)
			default_font = XLoadQueryFont (dpy, "fixed");
	}

	//-------------------------------------------------
	// In automatic mode, we set the size from the 
	// window manager's suggested values.
	//
	if (automatic) {
		Window root = XDefaultRootWindow (dpy);
		Atom name_atom, utf8_atom;
		Atom type;
		int format;
		unsigned long n_items, after;
		long *data = NULL;
		name_atom = XInternAtom (dpy, "_NET_WORKAREA", false);
		utf8_atom = XInternAtom (dpy, "UTF8_STRING", false);
		bool good = false;
		if (Success == XGetWindowProperty (dpy, root, name_atom, 0, 65536,
					false, utf8_atom, &type, &format,
					&n_items, &after, (unsigned char**) &data)) {
			if (data) {
				if (n_items >= 4) {
					requested_window_x = data [0];
					requested_window_y = data [1];
					requested_window_width = data [2];
					requested_window_height = data [3];
					good = true;
				}
				XFree (data);
			}
		} 

		if (requested_window_width > dpy_width)
			requested_window_width = dpy_width;
		if (requested_window_height > dpy_height)
			requested_window_height = dpy_height;
		border_size = 0;

		if (!good) 
			log_warning ("Unable to get window size/position from window manager.");
	}
}

void
App::shutdown ()
{
	quitnow = true;
}

App::App () 
{
	n_windows = 0;
	n_animations = 0;
        touch_focus_widget = NULL;

	//-----------------
	// Keyboard logic.
	//
	keyboard = NULL;
        key_focus_widget = NULL;
	alt_pressed = false;
	meta_pressed = false;

	//-----------------
	// Set up message Q.
	//
	n_indirect_messages = 0;
	for (int i = 0; i < MAX_INDIRECT_MESSAGES; i++)
		indirect_messages [i] = NULL;

        fg = RGB_BLACK;
	bg = RGB_WHITE;
	bg_gradient = RGB_GRAY;
	highlight_color = RGB_CYAN;
	selected_color = RGB_BLUE;
	border_color = RGB_NAVYBLUE;
	quitnow = false;
	program_name = "Unnamed App";

	styles_path [0] = 0;

	//---------------------------------------
	// Report the use of this widget system.
	//
	log ("This app is using TouchWidgets version "TOUCHWIDGETS_RELEASE".");

	//-------------------
	// Open the display.
	//
	char *dpy_str = getenv("DISPLAY");
	if (!dpy_str) 
		dpy_str = (char*) ":0.0";
	if (!(dpy = XOpenDisplay (dpy_str))) {
		log_fatal ("Cannot open the display.");
	}
	dpy_width = DisplayWidth (dpy, 0);
	dpy_height = DisplayHeight (dpy, 0);
	dpy_depth = DefaultDepth (dpy, 0);
	char tmp [100];
	sprintf (tmp, "Display is %dx%d, depth %d", dpy_width, dpy_height, dpy_depth);
	log (tmp);

	//-----------------------------
	// Set up the color allocation.
	//
	visual = DefaultVisual (dpy, 0);
	XVisualInfo vi;
	bzero (&vi, sizeof(XVisualInfo));
	vi.visual = visual;
	int i;
	XVisualInfo *vi2 = XGetVisualInfo (dpy, VisualNoMask, &vi, &i);
	if (!vi2) 
		log_fatal ("Unable to get X-Windows screen color mapping information.");
	switch (visual_type = get_visual_class (vi2)) {
	case TrueColor:
		log ("Display color model is TrueColor.");
		break;
	case DirectColor:
		log ("Display color model is DirectColor.");
		break;
	default:
		log_fatal ("TouchWidgets requires a TrueColor or DirectColor display.");
	}
}

App::~App ()
{
	if (keyboard) {
		delete keyboard;
		keyboard = NULL;
	}

	for (int i = n_indirect_messages - 1; i >= 0; i--) 
		delete indirect_messages [i];

	for (int i = n_windows-1; i >= 0; i--) 
		delete controllers [i];

	// XX Tell all controllers to prepare to shut down e.g. save user's data.

	if (default_title_font_bold) {
		XFreeFont (dpy, default_title_font_bold);
		default_title_font_bold = NULL;
	}
	if (default_font) {
		XFreeFont (dpy, default_font);
		default_font = NULL;
	}
	if (default_font_bold) {
		XFreeFont (dpy, default_font_bold);
		default_font_bold = NULL;
	}
	if (default_font_italic) {
		XFreeFont (dpy, default_font_italic);
		default_font_italic = NULL;
	}
	if (default_font_bold_italic) {
		XFreeFont (dpy, default_font_bold_italic);
		default_font_bold_italic = NULL;
	}

	// XX More to do...
}

void
App::log (const char *str)
{
	fprintf (stderr, "%s\n", str);
}

void
App::log_warning (const char *str)
{
	fprintf (stderr, "Warning: %s\n", str);
}

void
App::log_fatal (const char* str)
{
	fprintf (stderr, "Fatal error: %s\n", str);
	exit (1);
}

Widget*
App::identify_widget_at_absolute_point (Point p, Point *return_relative_point)
{
	Controller *c = get_top_controller ();
	if (!c)
		return NULL;

	Widget *rw = c->get_root_widget ();
	if (!rw)
		return NULL;

	Widget *w = rw->identify_child_for_absolute_point (p, return_relative_point);
	return w;
}

void 
App::add_animation (Widget* w)
{
	animations [n_animations++] = w;
}

void 
App::remove_animation (Widget* w)
{
	if (n_animations <= 1) {
		n_animations = 0;
		return;
	}

	unsigned ix = 0;
	while (animations [ix] != w) {
		ix++;
	}
	if (ix >= n_animations) {
		app->log_warning ("Widget not in animation list .'. cannot remove from list.");
		return;
	}

	while (ix < n_animations-1) {
		animations [ix] = animations [ix+1];
		ix++;
	}
	--n_animations;
}

void 
App::indirect_message_after_delay (MessageBundle *bundle, time_t delay)
{
	if (!bundle)
		return;

	if (n_indirect_messages >= MAX_INDIRECT_MESSAGES) {
		log_warning ("Too many indirect messages.");
		delete bundle;
		return;
	}

	bundle->fire_time = get_millisecond_time() + delay;
	indirect_messages [n_indirect_messages++] = bundle;
}

void 
App::indirect_message_after_delay (Object *sender, Object *recipient, Message m, MessageParameter param, time_t delay)
{
	if (!sender || !recipient)
		return;

	MessageBundle *bundle = new MessageBundle ();
	bundle->sender = sender;
	bundle->recipient = recipient;
	bundle->message = m;
	bundle->param = param;
	indirect_message_after_delay (bundle, delay);
}

void 
App::indirect_message (Object *sender, Object *recipient, Message m, MessageParameter param)
{
	indirect_message_after_delay (sender, recipient, m, param, 0);
}

uint32_t 
App::get_microsecond_time ()
{
        struct timeval t;
        gettimeofday (&t, NULL);
        return (uint32_t) (t.tv_sec * 1000000 + t.tv_usec);
}

uint32_t
App::get_millisecond_time ()
{
        struct timeval t;
        gettimeofday (&t, NULL);
        return (uint32_t) (t.tv_sec * 1000 + t.tv_usec / 1000);
}

void 
App::end_touch_speed_estimation ()
{
	for (int i = 0; i < TOUCH_DELTA_LOG_SIZE; i++)
        	touch_delta_speeds [i] = -9999;
        
        last_touch_time = 0;
	last_touch_point = Point_zero ();
}

void 
App::begin_touch_speed_estimation (uint32_t t, Point p)
{
	for (int i = 0; i < TOUCH_DELTA_LOG_SIZE; i++)
        	touch_delta_speeds [i] = -9999;
        
        last_touch_time = t;
	last_touch_point = p;
}

void 
App::update_touch_speed_estimation (uint32_t t, Point p)
{
	uint32_t dt = t - last_touch_time;
	last_touch_time = t;

	float distance = Point_distance (p, last_touch_point);

	last_touch_point = p;
	float speed = 0;
	if (dt > 0) {
		speed = distance / (float) dt;
	} else {
		speed = distance / .5;
	}

	for (int i = TOUCH_DELTA_LOG_SIZE-1; i > 0; i--)
		touch_delta_speeds [i] = touch_delta_speeds [i-1];

	touch_delta_speeds [0] = speed;

#ifdef DEBUG_MOMENTUM
	printf ("ADDED speed estimation:\n");
	printf ("\tdt = %lu\n",(unsigned long) dt);
	printf ("\tdistance = %lf\n", distance);
	printf ("\tspeed = %lf\n", speed );
#endif
}

float
App::get_touch_speed_estimation (void)
{
	float amount;
	float sum = 0;
	int n =0;
	int i = 0;
	while (i < TOUCH_DELTA_LOG_SIZE && (amount = touch_delta_speeds[i]) != -9999) {
		if (fabs (amount) > 0.01) {
			sum += amount;
			n++;
		}
		i++;
	}

	if (!n)
		return 0.2;

	float speed = sum / (float) n;
#ifdef DEBUG_MOMENTUM
	printf ("Returning speed estimation: %g\n", speed);
#endif
	return speed;
}

void 
App::begin_internal_animation (Widget *w)
{
        internally_animating_widget = w;
}

void 
App::end_internal_animation (Widget *w)
{
        if (internally_animating_widget == w) {
        	internally_animating_widget = NULL;
	}

	// XX Need to be able to trigger an event at end of internal animation.
}

#if 0
void 
App::associate_name_and_object (const char *name, Object* w)
{
	names.insert (std::make_pair(name, w));
}
#endif

void 
App::beep (unsigned frequency, unsigned duration_in_ms)
{
	if (!duration_in_ms)
		duration_in_ms = 1;
	else if (duration_in_ms > 5000)
		duration_in_ms = 5000;

	XKeyboardControl tone;
	tone.bell_pitch = frequency;
	tone.bell_duration = duration_in_ms;
	XChangeKeyboardControl (dpy, KBBellPitch | KBBellDuration, &tone);

	XBell (dpy, 0);

	// Restore defaults.
	tone.bell_pitch = -1;
	tone.bell_duration = -1;
	XChangeKeyboardControl (dpy, KBBellPitch | KBBellDuration, &tone);
}

void
App::hide_cursor ()
{
	if (n_windows < 1)
		return;

	char pixels [8] = { 0,0,0,0,0,0,0,0 };
	Pixmap pixmap;
	XColor color, exact;
	Cursor cursor;

	for (int i = 0; i < n_windows; i++) {
		Window window = windows [i];
		if (window) {
			pixmap = XCreateBitmapFromData (dpy, window, pixels, 8, 8);
			XAllocNamedColor (dpy, colormap, "white", &color, &exact);
			cursor = XCreatePixmapCursor (dpy, pixmap, pixmap, &color, &color, 0, 0);
			XDefineCursor (dpy, window, cursor);
			XFreeCursor (dpy, cursor);
		}
	}
}

bool 
App::parse_color (const char *name, RGB *rgb_return)
{
	if (!name || !rgb_return)
		return false;

	if ('#' == *name) {
		unsigned long rgb;
		if (1 == sscanf (name+1, "%lx", &rgb)) {
			*rgb_return = rgb;
			return true;
		}
		return false;
	}

	if ('0' == *name && 'x' == name[1]) {
		unsigned long rgb;
		if (1 == sscanf (name+2, "%lx", &rgb)) {
			*rgb_return = rgb;
			return true;
		}
		return false;
	}

	XColor color, exact;
	if (XAllocNamedColor (dpy, colormap, name, &color, &exact)) {
		*rgb_return = exact.pixel;
		return true;
	}

	return false;
}

static int
readline (FILE *f, char *line, int max)
{
	int ch=0, ix=0;

	if (!f || !line || max<2)
		return EOF;

	while (EOF != (ch = fgetc (f))) 
	{
		if (ch == '\r' || ch == '\n')
			break;
		if (ch == '#') {
			//--------------------
			// Skip past comments.
			//
			while (EOF != (ch = fgetc (f))) {
				if (ch == '\n' || ch == '\r')
					break;
			}
			break;
		}

		line [ix++] = ch;
		if (ix == max-1)
			break;
	}

	line [ix] = 0;

	if (EOF == ch && !ix)
		return ch;

	return ix;
}

void 
App::set_styles_path (const char *path)
{
	if (!path)
		return;
	strncpy (styles_path, path, PATH_MAX-1);
}

//-----------------------------------------------------------------------------
// Name:	fetch_style
// Purpose:	This is a simple parser for a file formatted thus:
// 		#objectName.property = value.
//			Example: #happyButton.text: Press Me. 
// 		className.property = value.
//			Example: .redButtons.borderColor: red.
// Note:	Returns allocated string that needs freeing.
//-----------------------------------------------------------------------------

const char *
App::fetch_style (const char *entity_name, const char *property_name)
{
	if (!property_name)
		return NULL;

	FILE *f = fopen (styles_path, "rb");
	if (!f)
		return NULL;

	if (!entity_name)
		entity_name = "*";

	char line [256];
	while (EOF != readline (f, line, 255)) {
		if (!line [0]) 
			continue;

		// Skip comments.
		if ('/' == line[0] && '/' == line[1])
			continue;

		char *start = line;
		while (*start && isspace ((int) *start))
			start++;
		char *property = strchr (start+1, '.');
		if (!property)
			continue;
		*property++ = 0;
		char *value = strchr (property, ':');
		if (!value)
			continue;
		*value++ = 0;
		char *name = line;

		// Truncate whitespace at start of data.
		//
		while (*name && isspace (*name))
			name++;
		while (*property && isspace (*property))
			property++;
		while (*value && isspace (*value))
			value++;

		// Truncate whitespace at end of data.
		//
		char *tmp;
		tmp = strchr (name, ' '); 
		if (tmp) *tmp = 0;
		tmp = strchr (name, '\t'); 
		if (tmp) *tmp = 0;

		tmp = strchr (property, ' '); 
		if (tmp) *tmp = 0;
		tmp = strchr (property, '\t'); 
		if (tmp) *tmp = 0;

		tmp = strchr (value, ' '); 
		if (tmp) *tmp = 0;
		tmp = strchr (value, '\t'); 
		if (tmp) *tmp = 0;

		// Look for the exact object name.
		//
		if (!strcasecmp (name, entity_name) && !strcasecmp (property_name, property)) {
			fclose (f);
			return strdup (value);
		}
	}

	fclose (f);
	return NULL;
}

//-----------------------------------------------------------------------------
// Translation logic. XX Move to its own class.
//-----------------------------------------------------------------------------

enum { 
	LANG_DE = 0,
	LANG_FR = 1,
	LANG_ES = 2,
	LANG_NL = 3,
	LANG_RU = 4,
};

const char *
App::translate (const char *str)
{
	char *l = getenv ("LANG");
	if (!l || strlen (l) < 2)
		return str;
	char lang [3];
	lang [0] = l [0];
	lang [1] = l [1];
	lang [2] = 0;
	
	if (!strcmp (lang, "en"))
		return str;

	// Identify the language.
	//
	int code = -1;
	if (!strcmp (lang, "de"))
		code = LANG_DE;
	else if (!strcmp (lang, "fr"))
		code = LANG_FR;
	else if (!strcmp (lang, "es"))
		code = LANG_ES;
	else if (!strcmp (lang, "nl"))
		code = LANG_NL;
	else if (!strcmp (lang, "ru"))
		code = LANG_RU;

	// XX Need to use a file.
	//
	if (code < 0)
		return str;
	if (!strcmp (str, "Untitled")) {
		static const char *str_Untitled [] = {
			"Unbetitelt",
			"Sans titre",
			"Intitulado",
			"Naamloos",
			"неназванный",
		};
		return str_Untitled [code];
	}
	return str;
}

//-----------------------------------------------------------------------------
// Popup logic.
//-----------------------------------------------------------------------------

bool 
App::present_keyboard (Controller* controller, Dimension keyboard_height)
{
	// XX In future, use animation.

	if (keyboard_height < 1)
		keyboard_height = DEFAULT_KEYBOARD_HEIGHT;

	push_window (WINDOW_KEYBOARD, controller, dpy_width, keyboard_height, RGB_RED, "Keyboard");

	return true;
}

bool 
App::dismiss_keyboard (void)
{
	// XX In future, use animation.

	if (!keyboard)
		return false;
	if (n_windows < 1)
		return false;
	if (window_types [n_windows - 1] != WINDOW_KEYBOARD)
		return false;

	pop_window ();
	return true;
}

bool 
App::present_popup (Controller* controller)
{
	int popup_width = 350;
	int popup_height = 217;

	push_window (WINDOW_ALERT, controller, popup_width, popup_height, RGB_RED, "Alert");

	return true;
}

bool 
App::dismiss_popup (void)
{
	pop_window ();

	return true;
}

void
App::keep_window_on_top (Window win)
{
}

void
App::receive_message (Object *sender, Message m, MessageParameter p)
{
	// NOTE: On-screen keyboard can be deleted indirectly.
	if (sender && sender == keyboard && !strcmp (m, DISMISS_KEYBOARD)) {
		dismiss_keyboard ();
	}

}

