/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/

#ifndef _TOUCHWIDGETS_APP_H
#define _TOUCHWIDGETS_APP_H

#include <list>
#include <map>
#include <queue>

#include <TouchWidgets/Common.h>
#include <TouchWidgets/Object.h>
#include <TouchWidgets/Widget.h>
#include <TouchWidgets/Controller.h>
#include <TouchWidgets/Message.h>
#include <TouchWidgets/Keyboard.h>

namespace TouchWidgets {

class MessageBundle 
{
public:
	Message message;
	MessageParameter param;
	Object *recipient;
	Object *sender;
	uint32_t fire_time;

	MessageBundle () {
		message = NULL;
		recipient = NULL;
		sender = NULL;
		param = 0;
		fire_time = 0;
	}
	~MessageBundle () { }
};

enum {
	WINDOW_MAIN = 0,
	WINDOW_ALERT = 1,
	WINDOW_KEYBOARD = 2,
	WINDOW_ROOT = 3,
};

class App : public Object
{
	int process_event (XEvent *e);

	bool quitnow; // true => Stop the event loop.

	// Trees of controllers.
	//Controller *root_controller;
	//Controller *popup_root_controller;

	//-----------------------
	// Current animations.
	unsigned n_animations;
	// std::list<Widget*> animations;
#define MAX_ANIMATIONS (256)
	Widget* animations[MAX_ANIMATIONS];

#define MAX_WINDOWS 10
	int n_windows;
	Window windows [MAX_WINDOWS];
	char window_types [MAX_WINDOWS];
	bool window_is_initialized [MAX_WINDOWS];
	Controller *controllers [MAX_WINDOWS];

	Widget *key_focus_widget;
	Widget *touch_focus_widget;
	bool alt_pressed;
	bool meta_pressed;

	//-----------------------
	// Message queue.
	//
#define MAX_INDIRECT_MESSAGES 256
	int n_indirect_messages;
	MessageBundle *indirect_messages [MAX_INDIRECT_MESSAGES];

	//-----------------------
	// Touch speed estimation.
	//
        void begin_touch_speed_estimation (uint32_t t, Point p);
        void update_touch_speed_estimation (uint32_t t, Point p);
        void end_touch_speed_estimation (void);
#define TOUCH_DELTA_LOG_SIZE 16
        float touch_delta_speeds [TOUCH_DELTA_LOG_SIZE];
        uint32_t last_touch_time;
        Point last_touch_point;

	Widget *internally_animating_widget;

	//-----------------------
	// Fetching of styles.
	//
	char styles_path [PATH_MAX];

	void keep_window_on_top (Window);

	Window get_top_window (void) { if (n_windows >= 1) return windows [n_windows-1]; else return None; }
	Controller *get_top_controller (void) { if (n_windows >= 1) return controllers [n_windows-1]; else return NULL; }
	Controller *get_root_controller_for_window (Window w);

	void pop_window ();

	Keyboard *keyboard;

public:
	Window get_window_for_root_controller (Controller *rc);

	void push_window (int type, Controller *c, int width, int height, RGB bg, const char *title);
	App (void);
	~App (void);

	virtual const char* class_name () { return "App"; }

	float get_touch_speed_estimation (void);

	const char *fetch_style (const char *object_name, const char *property_name);
	void set_styles_path (const char *path); // XX In future allow URL.

	long requested_window_width; 
	long requested_window_height;
	long requested_window_x; 
	long requested_window_y;

	const char *program_name;

	bool window_is_root;

	Window get_root_window (Widget*);
	void present_root_controller (Controller *controller);

	void initialize (int argc, char **argv,
			char *program_name,
			bool automatic,
			ushort window_w,
			ushort window_h,
			uint32_t bgcolor,
			ushort border_size,
			uint32_t border_color);
	void run (void);
	void shutdown (void);
	void hide_cursor (void);

	void beep (unsigned frequency, unsigned duration_in_ms);

	bool parse_color (const char *name, RGB *rgb_return);

	//----------
	// Popups.
	//
	bool present_popup (Controller*);
	bool dismiss_popup (void);

	//----------
	// Keyboards.
	//
	bool present_keyboard (Controller*, Dimension height);
	bool dismiss_keyboard (void);

	//----------
	// Animation.
	//
	void add_animation (Widget*);
	void remove_animation (Widget*);
	void begin_internal_animation (Widget*);
        void end_internal_animation (Widget*);

	//---------------------
	// Logging.
	//
	void log (const char*);
	void log_warning (const char*);
	void log_fatal (const char*);

	//-----------
	// Language.
	//
	const char *translate (const char*);

	//----------
	// X11 stuff
	//
	Display *dpy;
	uint16_t dpy_width, dpy_height, dpy_depth;
	//Window window;
	//Window popup_window; 
	XImage *view_image;
	GC window_text_gc;
	XFontStruct *default_title_font_bold;
	XFontStruct *default_font;
	XFontStruct *default_font_bold;
        XFontStruct *default_font_italic;
        XFontStruct *default_font_bold_italic;
	Visual *visual;
	int visual_type;
	Colormap colormap;

	Widget *identify_widget_at_absolute_point (Point p, Point *return_relative_point);

	void request_key_focus (Widget *);
	void release_key_focus (Widget *);
	Widget *get_key_focus_widget (void) { return key_focus_widget; }
	void convey_key_to_key_focus_widget (KeySym, bool down);

	//----------
	// Timing.
	//
	uint32_t get_microsecond_time ();
	uint32_t get_millisecond_time ();

	//------------------------
	// Application-wide colors.
	//
	RGB fg, bg, highlight_color, selected_color, border_color;
	RGB bg_gradient; // <-- Second color for gradients, which transition from bg (at top) to bg_gradient (at bottom).

	//--------------------
	// Delayed messages.
	//
	void indirect_message_after_delay (MessageBundle *bundle, time_t delay);
	void indirect_message_after_delay (Object *sender, Object *recipient, Message m, MessageParameter param, time_t delay);
	void indirect_message (Object *sender, Object *recipient, Message m, MessageParameter param);
        void receive_message (Object *sender, Message m, MessageParameter p);
};

}

extern TouchWidgets::App *app;

#endif

