/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <stdlib.h>
#include <X11/Xmu/Drawing.h>

#include <TouchWidgets/App.h>
#include <TouchWidgets/Button.h>

// #define DEBUG_BUTTON

using namespace TouchWidgets;

void
Button::setup_colors ()
{
	button_dark = RGB_GRAY_DARK;
	button_medium = RGB_GRAY_MEDIUM;
	button_medium2 = RGB_GRAY_MEDIUM2;
	button_light2 = RGB_GRAY_LIGHT2;
	button_light = RGB_GRAY_LIGHT;

	button_selected_color = RGB_YELLOW;
	button_selected_fg = RGB_RED;
	button_selected_dark = RGB_change_intensity (button_selected_color, -.2); 
	button_selected_medium2 = RGB_change_intensity (button_selected_color, -.1); 
	button_selected_medium = RGB_change_intensity (button_selected_color, -.15); 
	button_selected_light2 = RGB_change_intensity (button_selected_color, .3); 
	button_selected_light = RGB_change_intensity (button_selected_color, .6); 
}

Button::Button () 
{
	string = NULL;
	delegate = NULL;
	selected = false;
	use_bold_default_font = false;
	value = 0;
	border = 5;
	type = BUTTON_WIDGET_ROUND;
	// accelerator = sym;

	setup_colors ();
}

Button::Button (const char *s) 
{
	type = BUTTON_WIDGET_ROUND;
	delegate = NULL;
	selected = false;
	value = 0;
	border = 5;

	if (s)
        	string = strdup (s);
	else
        	string = strdup ("");

	setup_colors ();
}

Button::~Button ()
{
	if (string) {
		free ((void*) string);
		string = NULL;
	}
}

void
Button::draw (Rect r)
{
        if (!window || hidden)
                return;

	XGCValues gcv;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	GC gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (r, gc);

	int button_is_out = !value;

	Dimension w = rect.size.width;
	Dimension h = rect.size.height;

	if (!border) 
		border=1;

	XFontStruct *font = !use_bold_default_font ? app->default_font : app->default_font_bold;

	XSetFont (app->dpy, gc, font->fid);

	switch (type) {
	case BUTTON_WIDGET_PLAIN: {
		} break;

	case BUTTON_WIDGET_3D_RECT:
		draw_3d_border (r, button_is_out); // XX codereview: selected not supported.
		break;

	case BUTTON_WIDGET_ROUND: {
		int corner_radius = 10;

		if (parent) {
			// Clear the corner rectangles.
			//
			XSetForeground (app->dpy, gc, parent->bg);
			XFillRectangle (app->dpy, window, gc, 0, 0, w, corner_radius);
			XFillRectangle (app->dpy, window, gc, 0, h - corner_radius, w, corner_radius);
			XFillRectangle (app->dpy, window, gc, 0, corner_radius, corner_radius, h - 2*corner_radius);
			XFillRectangle (app->dpy, window, gc, w - corner_radius, corner_radius, 
				corner_radius, h - 2*corner_radius);
		}

		if (!selected)
			XSetForeground (app->dpy, gc, button_is_out ? button_medium : button_medium2);
		else
			XSetForeground (app->dpy, gc, button_is_out ? button_selected_medium : button_selected_medium2);
		XmuFillRoundedRectangle (app->dpy, window, gc, 0, 0, w-1, h-1, corner_radius, corner_radius);

		if (!selected)
			XSetForeground (app->dpy, gc, !button_is_out ? button_light2 : button_dark);
		else
			XSetForeground (app->dpy, gc, !button_is_out ? button_selected_light2 : button_selected_dark);
		XmuDrawRoundedRectangle (app->dpy, window, gc, 0, 0, w-1, h-1, corner_radius, corner_radius);
		} break;

	}

	XSetForeground (app->dpy, gc, !selected ? fg : button_selected_fg);

	// Get the text width(s) for up to 2 lines.
	//
	int align = TEXT_ALIGN_CENTER;
	int len = strlen (string);
	int ascent = font->ascent;
	Coordinate x = border;
	Dimension tw = 0;
	Dimension tw2 = 0;
	char *s2 = strchr ((char*) string, '\n');
	if (!s2) {
		tw = XTextWidth (font, string, len);
		switch (align) {
			case TEXT_ALIGN_RIGHT: 
				x += w - tw;
				x -= 2*border;
				break;
			case TEXT_ALIGN_CENTER: 
				x = border + ((w-2*border) - tw)/2;
				break;
		}

		int y = ascent + border + ((h-2*border) - ascent)/2;
		XDrawString (app->dpy, window, gc, x, y, string, len);
	} else {
		len = s2 - string;
		tw = XTextWidth (font, string, s2 - string);
		s2++;
		int len2 = strlen (s2);
		tw2 = XTextWidth (font, s2, strlen (s2));

		switch (align) {
			case TEXT_ALIGN_RIGHT: 
				x += w - tw;
				x -= 2*border;
				break;
			case TEXT_ALIGN_CENTER: 
				x = border + ((w-2*border) - tw)/2;
				break;
		}

		Coordinate x2 = border;
		switch (align) {
			case TEXT_ALIGN_RIGHT: 
				x2 += w - tw2;
				x2 -= 2*border;
				break;
			case TEXT_ALIGN_CENTER: 
				x2 = border + ((w-2*border) - tw2)/2;
				break;
		}

		// Draw the upper line.
		int line_height = ascent + font->descent;
		int y = border + (h-2*border)/2 - line_height + ascent;
		XDrawString (app->dpy, window, gc, x, y, string, len);

		// Draw the lower line.
		y = border + (h-2*border)/2 + ascent;
		XDrawString (app->dpy, window, gc, x2, y, s2, len2);
	}

	XFreeGC (app->dpy, gc);
}

void
Button::receive_message (Object *sender, Message m, MessageParameter tmp)
{
	if (!m) 
		return;
	
	if (!strcmp (m, TOUCH_DOWN_INSIDE)) {
		value = true;
		redraw ();
	}
	else if (!strcmp (m, TOUCH_UP_OUTSIDE)) {
		value = false;
		redraw ();
	}
	else if (!strcmp (m, TOUCH_UP_INSIDE)) {
		value = false;
		redraw ();

		if (delegate) {
#ifdef DEBUG_BUTTON
			puts ("Button: telling delegate about touch_up_inside.");
#endif
			send_message (delegate, TOUCH_UP_INSIDE);
		}
	}
}

