/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <stdlib.h>
#include <string.h>
#include <X11/Xmu/Drawing.h>

// #define DEBUG_BUTTONBAR

#include <TouchWidgets/App.h>
#include <TouchWidgets/ButtonBar.h>

using namespace TouchWidgets;

#define TWEAK 2
#define UNDERLINE_OFFSET 3

ButtonBar::ButtonBar () 
{
	delegate = NULL;
	n_buttons = 0;
	selection = -1;
	tap_down = false;
	border = DEFAULT_BUTTONBAR_BORDER;
	fg = DEFAULT_BUTTONBAR_FGCOLOR;
	bg = DEFAULT_BUTTONBAR_BGCOLOR;
	border_color = RGB_GRAY;

	int i;
	for (i = 0; i < MAX_BARBUTTONS; i++)
		strings[i] = NULL;

	// accelerator = sym;
	// XX Register with app for META-1, META-2 etc
}

ButtonBar::~ButtonBar ()
{
	if (n_buttons) {
		int i;
		for (i = 0; i < n_buttons; i++) {
			free (strings [i]);
		}
	}
}

void
ButtonBar::add_button (const char *s)
{
	if (s && n_buttons < MAX_BARBUTTONS) {
		strings [n_buttons++] = strdup (s);
	}
}

void
ButtonBar::draw (Rect clip)
{
        if (!window || hidden)
                return;

	XGCValues gcv;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	GC gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (clip, gc);

	Dimension w = rect.size.width;
	Dimension h = rect.size.height;

	if (border) 
		draw_3d_border (clip, false);

#ifdef DEBUG_BUTTONBAR
	printf ("ButtonBar: n_buttons = %d\n", n_buttons);
	printf ("ButtonBar: bgcolor = %x\n", (unsigned) bg);
#endif
	if (!n_buttons)
		return;

	XSetForeground (app->dpy, gc, fg);

	XFontStruct *font = app->default_font_bold;
	XSetFont (app->dpy, gc, font->fid);

	int ascent = font->ascent;
	int th = ascent + font->descent;
	Coordinate y = ascent + (h - border - th)/2 + TWEAK;

	int i;
	for (i = 0; i < n_buttons; i++) {
		char *s = strings [i];
		int tw = XTextWidth (font, s, strlen(s));

		Dimension button_width = (w-2*border) / n_buttons;
		Coordinate x0 = border + i * button_width;
		Coordinate x = x0 + (button_width - tw) / 2.;
		XDrawString (app->dpy, window, gc, x, y, s, strlen(s));

		if (i == selection) {
			XDrawLine (app->dpy, window, gc, x, y + UNDERLINE_OFFSET, x + tw, y + UNDERLINE_OFFSET);
		}
	}

	XFreeGC (app->dpy,gc);
}

void 
ButtonBar::select (int n)
{
	if (n < 0 || n >= n_buttons)
		deselect ();
	else {
		selection = n;
		redraw ();
	}
}

void 
ButtonBar::deselect (void)
{
	selection = -1;
	redraw ();
}

void
ButtonBar::receive_message (Object *sender, Message m, MessageParameter tmp)
{
	if (!m) 
		return;

	Point p = Point_from_message (tmp);
	
	if (!strcmp (m, TOUCH_DOWN_INSIDE)) {
		tap_down = true;
		redraw ();
	}
	else if (!strcmp (m, TOUCH_UP_OUTSIDE)) {
		tap_down = false;
		redraw ();
	}
	else if (!strcmp (m, TOUCH_UP_INSIDE)) {
		tap_down = false;

		if (delegate) {
#ifdef DEBUG_BUTTONBAR
			puts ("ButtonBar: telling delegate about touch_up_inside.");
#endif
			if (n_buttons) {
				Dimension button_width = (rect.size.width - 2*border) / n_buttons;
				int which = (p.x - border) / button_width;
				if (which < 0)
					which = 0;
				if (which >= n_buttons)
					which = n_buttons - 1;
				send_message (delegate, BUTTONBAR_SELECTION, which);
				selection = which;
			} else {
				send_message (delegate, BUTTONBAR_SELECTION, -1);
				selection = -1;
			}
		}
		redraw ();
	}
}

