/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <TouchWidgets/App.h>
#include <TouchWidgets/Canvas.h>

using namespace TouchWidgets;

void
Canvas::draw (Rect clip)
{
        if (!window || hidden)
                return;

	if (!pixels || !ximage)
		return;

	XGCValues gcv;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	GC gc = XCreateGC (app->dpy, window, 0, &gcv);
	XRectangle rr;
	int x = rr.x = clip.location.x;
	int y = rr.y = clip.location.y;
	int w = rr.width = clip.size.width;
	int h = rr.height = clip.size.height;
	XSetClipRectangles (app->dpy, gc, 0, 0, &rr, 1, Unsorted);

	int x2 = x + w;
	if (x2 > image_width)
		w -= (x2 - image_width);
	int y2 = y + h;
	if (y2 > image_height)
		h -= (y2 - image_height);

	XPutImage (app->dpy, window, gc, ximage, x, y, x, y, w, h);

	XFreeGC (app->dpy, gc);
}

void
Canvas::allocate_pixels ()
{
	if (pixels)
		free (pixels);
	pixels = NULL;

	if (image_width < 1 || image_height < 1) {
		app->log_warning ("Display has invalid dimensions.");
		return;
	}

	unsigned bytes_per_pixel = 3;
	image_ncomponents = 3;
	unsigned size = image_width * image_height * bytes_per_pixel;
	pixels = (uint8_t*) calloc (1, size);
	if (!pixels) {
		app->log_warning ("Cannot allocate pixels for image.");
		return;
	}

	if (ximage)
		XDestroyImage (ximage);
		
	ximage = (XImage*) malloc (sizeof(XImage));
	if (!ximage) {
		free (pixels);
		pixels = NULL;
		app->log_warning ("Cannot allocate image.");
		return;
	}

	// Set up an RGB XImage.
	//
	ximage->width = image_width;
	ximage->height = image_height;
	ximage->data = (char*) pixels;
	ximage->xoffset = 0;
	ximage->format = ZPixmap;
	ximage->bitmap_unit = 8;
	ximage->bitmap_pad = 8;
	ximage->depth = 24;
	ximage->byte_order = LSBFirst;
	ximage->bitmap_bit_order = LSBFirst;
	ximage->bits_per_pixel = 24;
	ximage->bytes_per_line = 3 * image_width;
	ximage->red_mask = 0xff0000;
	ximage->blue_mask = 0xff00;
	ximage->green_mask = 0xff;
	
	if (!XInitImage (ximage)) {
		app->log_warning ("Unable to init XImage.");
		return;
	}

	clear ();
}

Canvas::Canvas () 
{
	delegate = NULL;
	image_height = 0;
	image_width = 0;
	image_ncomponents = 0;
	pixels = NULL;
	ximage = NULL;
	bg = RGB_BLACK;
}

Canvas::Canvas (Size size, RGB color) 
{
	image_height = 0;
	image_width = 0;
	image_ncomponents = 0;
	pixels = NULL;
	ximage = NULL;

	if (size.width >= 1)
		image_width = size.width;
	if (size.height >= 1)
		image_height = size.height;
	if (image_width && image_height)
		allocate_pixels ();

	bg = color;
}

void
Canvas::setup (int w, int h, RGB color) 
{
	if (w < 1 || h < 1)
		return;

	if (pixels) {
		// app->log_warning ("Canvas: size already set.");
		return;
	}

	bg = color;
	image_width = w;
	image_height = h;
	allocate_pixels ();
}

void 
Canvas::clear (void)
{
	int n_pixels = image_width * image_height;
	if (!n_pixels || !pixels)
		return;

	uint8_t r = 255 & (bg >> 16);
	uint8_t g = 255 & (bg >> 8);
	uint8_t b = 255 & bg;
	uint8_t *p = pixels;

	// XX To do, speedup: write 12 bytes at a time.

	for (int i = 0; i < n_pixels; i++) {
		*p++ = b;
		*p++ = g;
		*p++ = r;
	}
}

RGB 
Canvas::get_pixel (Point p)
{
	if (!pixels || p.x < 0 || p.y < 0 || p.x >= image_width || p.y >= image_height)
		return bg;

	int x = (int) p.x;
	int y = (int) p.y;
	int ix = 3 * (y * image_width + x);
	uint32_t value = pixels [2+ix];
	value <<= 8;
	value |= pixels [1+ix];
	value <<= 8;
	value |= pixels [ix];
	return value;
}

void 
Canvas::put_pixel (Point p, RGB color)
{
	if (!pixels || p.x < 0 || p.y < 0 || p.x >= image_width || p.y >= image_height)
		return;

	int x = (int) p.x;
	int y = (int) p.y;
	int ix = 3 * (y * image_width + x);
	
	pixels [ix] = color;
	color >>= 8;
	pixels [ix+1] = color;
	color >>= 8;
	pixels [ix+2] = color;
}

void 
Canvas::draw_hline (int x, int y, int width, RGB color)
{
	if (!pixels || y < 0 || x >= image_width || y >= image_height)
		return;
	int x2 = x + width - 1;
	if (x2 < 0)
		return;
	if (x < 0) {
		width += x;
		x = 0;
	}
	if (x2 >= image_width) {
		width -= (x2 - image_width) + 1;
	}

	int ix = 3 * (y * image_width + x);
	uint8_t r = 255 & (color >> 16);
	uint8_t g = 255 & (color >> 8);
	uint8_t b = 255 & color;
	uint8_t *p = &pixels[ix];
	while (width-- > 0) {
		*p++ = b;
		*p++ = g;
		*p++ = r;
	}
}

void 
Canvas::draw_line (Point, Point, RGB)
{
}

void 
Canvas::draw_rect (Rect, RGB)
{
}

void 
Canvas::fill_rect (Rect r, RGB color)
{
	int x = (int) r.location.x;
	int y = (int) r.location.y;

	if (!pixels || x >= image_width || y >= image_height)
		return;
	int w = (int) r.size.width;
	int h = (int) r.size.height;

	if (x + w <= 0 || y + h <= 0)
		return;

	if (x < 0) {
		w += x;
		x = 0;
	}
	if (y < 0) {
		h += y;
		y = 0;
	}

	while (h > 0 && y < image_height) {
		draw_hline (x, y, w, color);
		h--;
		y++;
	}
}


