/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/

#ifndef _TOUCHWIDGETS_COMMON_H
#define _TOUCHWIDGETS_COMMON_H

#define TOUCHWIDGETS_RELEASE "0.21"

extern "C" {
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define XK_MISCELLANY
#include <X11/keysymdef.h>

#ifdef __linux__
#include <linux/limits.h> // PATH_MAX
#endif
}

// #define DEBUG_SCROLLING
// #define DEBUG_EVENT_LOOP
// #define DEBUG_MOMENTUM
// #define DEBUG_CREATION
// #define DEBUG_ANIMATION

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#if 1
#define ENTRY { fprintf (stderr, "Entered %s.\n", __PRETTY_FUNCTION__); fflush (NULL); }
#else
#define ENTRY 
#endif

typedef uint32_t RGB;

#define STRING_MILLISECONDS_PER_EVENT_LOOP_ITERATION "33.3"
#define MICROSECONDS_PER_EVENT_LOOP_ITERATION (33333) // 30 Hz
#define ANIMATION_SLICE (1) // # of event loop interations between an animation slice.
#define ANIMATION_SLICES_PER_SECOND ((1000000/MICROSECONDS_PER_EVENT_LOOP_ITERATION)/ANIMATION_SLICE)
#define LIMIT_OF_HUMAN_PERCEPTION (1./25.)

#define DEFAULT_KEYBOARD_HEIGHT (350)

// Not used.
#define DEFAULT_WINDOW_WIDTH (800)
#define DEFAULT_WINDOW_HEIGHT (400)

enum {
	TEXT_ALIGN_NONE = 0,	// Not specified.
	TEXT_ALIGN_LEFT = 1,	
	TEXT_ALIGN_RIGHT = 2,
	TEXT_ALIGN_CENTER = 3,
};

enum {
        WIDGET_ALIGN_TOP_LEFT = 0,
        WIDGET_ALIGN_MIDDLE_LEFT = 1,
        WIDGET_ALIGN_BOTTOM_LEFT = 2,
        WIDGET_ALIGN_TOP_CENTER = 3,
        WIDGET_ALIGN_MIDDLE_CENTER = 4,
        WIDGET_ALIGN_BOTTOM_CENTER = 5,
        WIDGET_ALIGN_TOP_RIGHT = 6,
        WIDGET_ALIGN_MIDDLE_RIGHT = 7,
        WIDGET_ALIGN_BOTTOM_RIGHT = 8,
        WIDGET_ALIGN_STRETCH = 9,
};
#define WIDGET_ALIGN_CENTER (WIDGET_ALIGN_MIDDLE_CENTER)

#define RGB_AMBER (0xffbf00)
#define RGB_BLACK (0)
#define RGB_BLUE (0xff)
#define RGB_CYAN (0xffff)
#define RGB_DARKGREEN (0x6400)
#define RGB_DARKORANGE (0xff8c00)
#define RGB_GRAY (0xe0e0e0)
#define RGB_GRAY_DARK 0x707070
#define RGB_GRAY_LIGHT 0xe0e0e0
#define RGB_GRAY_LIGHT2 0xb0b0b0
#define RGB_GRAY_MEDIUM  0xf0f0f0
#define RGB_GRAY_MEDIUM2 0xe0e0e0
#define RGB_GREEN (0x127000)
#define RGB_LIGHTGRAY (0xf0f0f0)
#define RGB_LIME (0xff00)
#define RGB_MAGENTA (0xff00ff)
#define RGB_MAROON (0x800000)
#define RGB_NAVYBLUE (0x80)
#define RGB_ORANGE (0xfc6b00)
#define RGB_PURPLE (0xa020f0)
#define RGB_RED (0xff0000)
#define RGB_STEELBLUE 0x6682b4
#define RGB_WHITE (0xffffff)
#define RGB_YELLOW (0xffff00)

typedef float Coordinate;
typedef float Dimension;

typedef struct {
	Coordinate x, y;
} Point;

static inline Point Point_new (Coordinate x, Coordinate y) {
	Point p;
	p.x = x;
	p.y = y;
	return p;
}

static inline Point Point_zero (void) {
	Point p = { 0, 0 };
	return p;
}

extern float Point_distance (Point p, Point p2);

typedef struct {
	Dimension width, height;
} Size;

typedef struct {
	Point location;
	Size size;
} Rect;

typedef struct {
	Dimension top;
	Dimension bottom;
	Dimension left;
	Dimension right;
} Margins; // Also for use with padding, border, insets.

static inline Margins Margins_new (Dimension t, Dimension b, Dimension l, Dimension r) {
	Margins m;
	m.top = t;
	m.bottom = b;
	m.left = l;
	m.right = r;
	return m;
}

static inline Size Size_new (Dimension w, Dimension h) {
	Size s;
	s.width = w;
	s.height = h;
	return s;
}

static inline Size Size_zero (void) {
	Size s = { 0, 0 };
	return s;
}

static inline Rect Rect_new (Coordinate x, Coordinate y, Dimension w, Dimension h) {
	Rect r;
	r.location.x = x;
	r.location.y = y;
	r.size.width = w;
	r.size.height = h;
	return r;
}

static inline Rect Rect_zero (void) {
	Rect r = {{ 0, 0 }, {0,0}};
	return r;
}

extern uint32_t getcolor (uint32_t rgb);
extern void freecolor(uint32_t p);

extern const char *Point_string (Point);
extern const char *Rect_string (Rect);
extern const char *Size_string (Size);

static inline bool Rect_equal (Rect r1, Rect r2) {
	return r1.location.x == r2.location.x && r1.location.y == r1.location.y && r1.size.width == r2.size.width && r1.size.height == r2.size.height;
}

extern RGB RGB_change_intensity (RGB color, float offset);

#define MIN2(AA,BB) ( (AA)<(BB) ? (AA) : (BB) )
#define MAX3(AA,BB,CC) ( MAX2(AA,BB) > (CC) ? MAX2(AA,BB) : (CC))
#define MAX2(AA,BB) ( (AA)>(BB) ? (AA) : (BB) )
#define MIN3(AA,BB,CC) ( MIN2(AA,BB) < (CC) ? MIN2(AA,BB) : (CC))

extern KeySym get_shifted_keysym (KeySym key);

#endif

