/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <stdlib.h>

#include <TouchWidgets/App.h>
#include <TouchWidgets/Container.h>

using namespace TouchWidgets;

#define TC_MODE_NONE 0
#define TC_MODE_DRAGGING 1

Container::Container () 
{
	dont_parse_touches = false;
	container_handles_taps = false;
	starting_point = Point_zero ();
	content_size = Size_zero ();
	mode = TC_MODE_NONE;
}

Container::~Container ()
{
	destroy_window ();
}

void 
Container::destroy_window (void)
{
	ENTRY
        for (int i = 0; i < n_children; i++) {
                Widget *child = get_nth_child (i);
                child->destroy_window ();
        }

        if (window) {
#ifdef DEBUG_CREATION
        printf ("Destroying window of %s. It was %x\n", name, (int) window);
#endif
                XDestroyWindow (app->dpy, window);
                window = None;
        }
}

bool
Container::get_dont_parse_touches ()
{
	return dont_parse_touches;
}

void
Container::set_dont_parse_touches (bool value)
{
	dont_parse_touches = value;
}

void 
Container::touches_start (Point* points, int n_points)
{
	if (!n_points)
		return;

	starting_point = points[0];
	mode = TC_MODE_NONE;
}

void 
Container::touches_continue (Point* points, int n_points)
{
	if (!n_points)
		return;

	Point p = points[0];
	if (mode == TC_MODE_DRAGGING) {
		Point delta = Point_new (p.x - starting_point.x, p.y - starting_point.y);
		scroll_continue (delta);
	} 
	else {
		double d = Point_distance (p, starting_point);
#define DRAG_THRESHOLD (10)
		if (d >= DRAG_THRESHOLD) {
			mode = TC_MODE_DRAGGING;

			scroll_start (Point_zero ());
			scroll_continue (Point_new (p.x - starting_point.x, p.y - starting_point.y));
		}
	}
}

void 
Container::touches_end (Point* points, int n_points)
{
	if (!n_points)
		return;

	Point p = points[0];

	if (mode == TC_MODE_DRAGGING) {
		scroll_end (Point_new (p.x - starting_point.x, p.y - starting_point.y));
	}
	else {
		double d = Point_distance (p, starting_point);
		if (d >= DRAG_THRESHOLD) {
			// This is the unusual case where we don't receive
			// the continued-motion event.
			//
			scroll_start (Point_zero ());
			scroll_end (Point_new (p.x - starting_point.x, p.y - starting_point.y));
		} else {
			// Here, determine which subwidget should receive
			// a tap event.
			//
#ifdef DEBUG_CONTAINER
			app->log ("Tap within Container.");
#endif

			Point return_point;
			Widget *w = identify_child_at_relative_point (p, &return_point);
			if (w) {
#ifdef DEBUG_CONTAINER
				printf ("It was widget %s that was tapped, cht=%d\n", w->name, container_handles_taps);
#endif
				if (!container_handles_taps) {
					Rect r = w->get_rect ();
					return_point.x -= r.location.x;
					return_point.y -= r.location.y;
					send_message (w, TOUCH_UP_INSIDE, MessageParameter_from_point (return_point));
				} else {
					int which = 0;
					while (which < n_children && children[which] != w)
						which++;
					if (which < n_children) {
						child_was_tapped (which);
					} 
				}
			} else {
#ifdef DEBUG_CONTAINER
				printf ("Cannot identify widget at point %s\n", Point_string (p));
#endif
			}
		}
	}
}

void
Container::scroll_start (Point p)
{
	printf ("Scrolling starts, offset = %s\n", Point_string (p));
}

void
Container::scroll_continue (Point p)
{
	printf ("Scrolling continues, offset = %s\n", Point_string (p));
}

void
Container::scroll_end (Point p)
{
	printf ("Scrolling ends, offset = %s\n", Point_string (p));
}

void 
Container::detect_content_size ()
{
	// RULE: Assume all widgets are right and below the (0,0) point.

	Size size = Size_zero ();
	if (n_children) {
		for (int i = 0; i < n_children; i++) {
			Widget *w = children[i];
			Rect r = w->get_rect ();
			Dimension ww = r.location.x + r.size.width;
			Dimension hh = r.location.y + r.size.height;
			if (ww > size.width)
				size.width = ww;
			if (hh > size.height)
				size.height = hh;
		}
	}
	set_content_size (size);
}

void 
Container::set_content_size (Size size)
{
	if (size.width < 1 || size.height < 1)
		return;

	content_size = size;
}

void
Container::child_was_tapped (int which)
{
        ENTRY
}

