/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <TouchWidgets/App.h>
#include <TouchWidgets/Controller.h>
#include <TouchWidgets/Container.h>

using namespace TouchWidgets;

#define DEFAULT_MAX_CHILDREN (16)

Controller::Controller () 
{
	delegate = NULL;
	max_children = 0;
	n_children = 0;
	children = NULL;
	next = NULL;
	parent = NULL;
	root_widget = NULL;
	title = NULL;
	size = Size_zero();
}

Controller::Controller (Widget *root) 
{
	delegate = NULL;
	max_children = 0;
	n_children = 0;
	children = NULL;
	next = NULL;
	parent = NULL;
	title = NULL;
	size = Size_zero();
	root_widget = NULL;

	if (root->get_window ()) {
		// No need for the widget to have a window.
		app->log_warning ("Widget already has a window, so cannot set it as root.");
	}
	else
		root_widget = root;
}

Controller::~Controller ()
{
ENTRY
        if (children) {
                for (int i = 0; i < n_children; i++) {
                        delete children[i];
                }
                free (children);
                children = NULL;
        }

	if (root_widget) {
#ifdef DEBUG_CREATION
		printf ("REMOVING root_widget %s from controller %s\n", root_widget->name, name);
#endif
		root_widget->remove_from_parent ();
		delete root_widget;
		root_widget = NULL;
	}
}

Widget *
Controller::get_root_widget (void)
{
	return root_widget;
}

#if 0
void 
Controller::set_root_widget (Widget* widget)
{
	if (root_widget) {
		delete root_widget;
		root_widget = NULL;
	}

	root_widget = widget;
}
#endif

Window
Controller::get_root_window (void)
{
	if (!root_widget)
		return 0;
	else
		return root_widget->get_window ();
}

void 
Controller::destroy_widgets (void)
{
	// Deallocate all of the widgets for this controller.
	// Destroy any windows.

	if (root_widget) {
		delete root_widget;
		root_widget = NULL;
	}
}

void 
Controller::layout_widgets (void)
{
}

static void 
expose_child_window (Widget *w, Rect r)
{
	if (!w) return;

	Rect wr = w->get_rect ();
	Rect r2 = r;
	r2.location.x -= wr.location.x;
	r2.location.y -= wr.location.y;

	// If the relative expose rectangle is outside
	// if this widget's bounds, we're done.
	//
	if (r2.location.x >= r2.size.width || r2.location.y >= r2.size.height 
	    || r2.location.x + r2.size.width <= 0 || r2.location.y + r2.size.height <= 0)
		return;

	w->draw (r2);

	for (int i = 0; i < w->n_children; i++) {
		Widget *w2 = w->get_nth_child (i); 
		expose_child_window (w2, r2);
	}
}

void 
Controller::expose_widgets (Rect r)
{
	if (!root_widget || root_widget->get_hidden ())
		return;

#ifdef DEBUG_EXPOSE
	printf ("In Controller::expose_widgets: exposure rect is %s\n", Rect_string(r));
#endif
	// Because r is absolute, make it relative.
	Rect rr = root_widget->get_rect ();
	r.location.x -= rr.location.x;
	r.location.y -= rr.location.y;

	expose_child_window (root_widget, r);
}

static Widget *
find_child_with_window (Widget *w, Window win)
{
	if (!w || !win) 
		return NULL;
	if (win == w->get_window())
		return w;

	if (!w->n_children)
		return NULL;

	for (int i = 0; i < w->n_children; i++) {
		Widget *w2 = find_child_with_window (w->get_nth_child (i), win);
		if (w2)
			return w2;
	}
	
	return NULL;
}

Widget *
Controller::get_widget_for_window (Window win)
{
	if (!root_widget)
		return NULL;

	return find_child_with_window (root_widget, win);
}

void 
Controller::set_rect (Rect r) 
{
	if (root_widget) {
		root_widget->set_rect (r);
		layout_widgets ();
	}
}

void 
Controller::will_appear (void)
{
}

void 
Controller::did_appear (void)
{
}

void
Controller::create_windows (void)
{
	if (root_widget) {
               	root_widget->set_hidden (false);
		Window w = root_widget->get_window ();
		if (!w) {
			Widget *p = root_widget->get_parent ();
			if (p)
				w = p->get_window ();
			else
				w = app->get_window_for_root_controller (this);

			root_widget->create_windows_for_self_and_children (w);
		}

		// RULE: Don't tell child controller will appear
		// because we don't know which one(s) this controller
		// will make visible.
	} else {
		app->log_warning ("Controller has no root widget.");
	}
}

void 
Controller::will_disappear (void)
{
}

void 
Controller::did_disappear (void)
{
}

void
Controller::destroy_windows (void)
{
	if (root_widget) {
                root_widget->set_hidden (true);

		if (n_children) {
			int i;
			for (i = 0; i < n_children; i++) {
				Controller *child = children [i];
				child->will_disappear ();
			}
		}

		root_widget->destroy_window ();
	}
}

void 
Controller::reallocate_children (int n)
{
	if (n < 4)
		n = 4;
	if (n <= max_children)
		n = 2*max_children;

	int new_max_children = n;
	children = (Controller**) realloc (children, sizeof(Controller*) * new_max_children);

	if (!children)
		app->log_fatal ("Out of memory.");

	for (unsigned i = n_children; i < new_max_children; i++)
		children [i] = NULL;

	max_children = new_max_children;
}

//-----------------------------------------------------------------------------
// Name:	add_child
// Purpose:	Add a child controller.
// Note:	Merely adding a child controller does *NOT* mean that its 
//		widgets (or windows thereof) need to be created.
//-----------------------------------------------------------------------------

void 
Controller::add_child (Controller* child)
{
	if (!children)
		reallocate_children (DEFAULT_MAX_CHILDREN);

	// Check whether controller is already a child.
        //
        int i;
        for (i = 0; i < n_children; i++) {
                if (child == children [i]) {
                        return;
                }
        }

	// RULE: If children array is full, increase
	// its size by 50%.
	//
	if (n_children && n_children >= max_children)
		reallocate_children (max_children + max_children/2);

	children [n_children++] = child;
	child->parent = this;

	Widget *w = child->get_root_widget ();
	if (w) {
		w->set_hidden (true);
	}
}

void 
Controller::remove_child (Controller *child)
{
	if (!children || !n_children || !child)
		return;

	int ix = -1;
	int i = 0;
	while (i < n_children) {
		if (children [i] == child) {
			ix = i;
			break;
		}
		i++;
	}
	if (ix < 0) {
		app->log_warning ("Attempt to remove child widget that was not found.");
		return;
	}

	int limit = n_children-1;
	while (ix < limit) {
		children[ix] = children[ix+1];
		ix++;
	}
	n_children--;

#if 0
	if (child->root_widget) 
		child->root_widget->destroy_windows_of_self_and_children ();
#endif
	child->will_disappear (); 
}

Controller* 
Controller::get_nth_child (int n)
{
	if (!children)
		return NULL;
	if (n < 0 || n >= n_children)
		return NULL;
	//==========

	return children [n];
}

#if 0
//-----------------------------------------------------------------------------
// Name:	present_child_at
// Purpose:	Prepare a child controller to be initialized and displayed.
// Note:	For modal controllers.
//-----------------------------------------------------------------------------

void 
Controller::present_child_at (int which, Rect r)
{
	if (!n_children || which < 0 || which > n_children || !children)
		return;

	if (!root_widget) {
		app->log_warning ("Cannot present child controller because parent does not have root widget.");
		return;
	}

	Controller *child = children [which];

	if (!child->root_widget) {
		app->log_warning ("Child controller forgot to set its root_widget.");
		return;
	}
	if (!child->root_widget->get_parent ()) {
		app->log_warning ("Child controller forgot to attach its root_widget.");
		return;
	}

	child->root_widget->set_rect (r);
	child->layout_widgets ();
	child->will_appear ();
}

//-----------------------------------------------------------------------------
// Name:	dismiss_child
// Purpose:	Remove and destroy a child.
// Note:	For modal controllers.
//-----------------------------------------------------------------------------

void 
Controller::dismiss_child (int which)
{
	if (!n_children || which < 0 || which > n_children || !children)
		return;

	Controller *child = children [which];

	Widget *root = child->root_widget;
	if (root)
		root->remove_from_parent ();

	delete child;
}
#endif

