/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <TouchWidgets/App.h>
#include <TouchWidgets/Graph.h>

using namespace TouchWidgets;

#define DEFAULT_LEFT_MARGIN (15)
#define DEFAULT_BOTTOM_MARGIN (15)
#define DEFAULT_TOP_MARGIN (5)
#define DEFAULT_RIGHT_MARGIN (5)

#ifndef MAXFLOAT
#define MAXFLOAT (1e37)
#endif

// Graph: useful for creating for instance a histogram.
//
void
Graph::draw (Rect r)
{
        if (!window || hidden)
                return;

	XGCValues gcv;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	GC gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (r, gc);

	int w = rect.size.width;
	int h = rect.size.height;

	determine_min_max ();

	XSetForeground (app->dpy, gc, bg);
	XFillRectangle (app->dpy, window, gc, 0, 0, w, h);
//	XClearWindow (app->dpy, window);

	XSetForeground (app->dpy, gc, fg);
	XDrawLine (app->dpy, window, gc, margins.left, h - margins.bottom, w - margins.right, h - margins.bottom);
	XDrawLine (app->dpy, window, gc, margins.left, h - margins.bottom - 1, margins.left, margins.top);

	switch (mode) {
	case GRAPH_MODE_BAR: 
		draw_bar_graph (gc);
		break;
	case GRAPH_MODE_LINE:
        	draw_line_graph (gc);
		break;
	case GRAPH_MODE_SCATTER:
        	draw_scatter_plot (gc);
		break;
	}

	XFreeGC (app->dpy, gc);
}

void
Graph::draw_bar_graph (GC gc)
{
	int w = rect.size.width;
	int h = rect.size.height;
	int n_data = yvalues.size ();
	if (!n_data)
		return;

	Dimension available_width = w - margins.left - margins.right - 1;
	Dimension available_height = h - margins.top - margins.bottom - 1;
	Dimension cell_width = available_width / n_data;

	// XX Need to implement proper scaling.

	for (int i = 0; i < n_data; i++) {
		float value = yvalues[i];
		RGB cell_color = colors[i];
		Coordinate x = margins.left + 1 + i* available_width / n_data;
		Coordinate y = h - margins.bottom - 1 - value;
		XSetForeground (app->dpy, gc, cell_color);
		XFillRectangle (app->dpy, window, gc, (int) x, (int) y, (int) (cell_width+.5), (int) (value+.5));
		x += cell_width;
	}
}

void
Graph::draw_line_graph (GC gc)
{
	int w = rect.size.width;
	int h = rect.size.height;

	int n_data = yvalues.size ();
	if (!n_data)
		return;

	Dimension available_width = w - margins.left - margins.right - 1;
	Dimension available_height = h - margins.top - margins.bottom - 1;

	RGB line_color = colors[0];

	if (n_data == 1) {
		float value = yvalues [0];
		Coordinate y = available_height * value / max_y;
		y = h - margins.bottom - 1 - y;
		XSetForeground (app->dpy, gc, line_color);
		XDrawLine (app->dpy, window, gc, margins.left+1, (int) y, w-1, (int) y);
		return;
	}

	Coordinate prev_x, prev_y;
	bool have_previous = false;

	// XX Need to implement proper scaling.

	for (int i=0; i < n_data; i++) {
		float value = yvalues [i];
		Coordinate x = margins.left + 1 + i * available_width / (n_data - 1);
		Coordinate y = available_height * value / max_y;
		y = h - margins.bottom - 1 - y;
		XSetForeground (app->dpy, gc, line_color);
		if (have_previous)
			XDrawLine (app->dpy, window, gc, (int) x, (int) y, (int) prev_x, (int) prev_y);
		prev_x = x;
		prev_y = y;
		have_previous = true;
	}
}

void
Graph::determine_min_max ()
{
	max_x = -MAXFLOAT;
	min_x = MAXFLOAT;
	max_y = -MAXFLOAT;
	min_y = MAXFLOAT;

	int n_data = yvalues.size ();

	for (int i = 0; i < n_data; i++) {
		float value;

		value = xvalues [i];
		if (value < min_x)
			min_x = value;
		if (value > max_x)
			max_x = value;

		value = yvalues [i];
		if (value < min_y)
			min_y = value;
		if (value > max_y)
			max_y = value;
	}
}

void
Graph::draw_scatter_plot (GC gc)
{
	int w = rect.size.width;
	int h = rect.size.height;

	int n_data = yvalues.size ();
	if (!n_data)
		return;

	Dimension available_width = w - margins.left - margins.right - 1;
	Dimension available_height = h - margins.top - margins.bottom - 1;

	for (int i=0; i < n_data; i++) {
		float xvalue = xvalues [i];
		float yvalue = yvalues [i];
		RGB color = colors[i];
		Coordinate x = margins.left + 1 + xvalue / max_x * available_width;
		Coordinate y = h - margins.bottom - 1 - yvalue / max_y * available_height;
		XSetForeground (app->dpy, gc, color);
		XFillRectangle (app->dpy, window, gc, (int) x-2, (int) y-2, 5, 5);
	}
}

Graph::~Graph ()
{
	clear_data ();
}

Graph::Graph () 
{
	delegate = NULL;
	mode = GRAPH_MODE_BAR;
	margins = Margins_new (DEFAULT_TOP_MARGIN, DEFAULT_BOTTOM_MARGIN, 
		DEFAULT_LEFT_MARGIN, DEFAULT_RIGHT_MARGIN);
	rect = Rect_zero ();
	fg = RGB_BLACK;
	bg = RGB_WHITE;
	color = RGB_GRAY;
}

void 
Graph::clear_data (void)
{
	xvalues.empty ();
	yvalues.empty ();
	colors.empty ();
}

void 
Graph::add_value (float value)
{
	xvalues.push_back (0);
	yvalues.push_back (value);
	colors.push_back (color);
}

void 
Graph::add_point (Point p)
{
	xvalues.push_back (p.x);
	yvalues.push_back (p.y);
	colors.push_back (color);
}

