/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <setjmp.h>

#include <TouchWidgets/App.h>
#include <TouchWidgets/ImageView.h>

using namespace TouchWidgets;

void
ImageView::draw (Rect clip)
{
        if (!window || hidden)
                return;

	// set_backing_store (true);

	if (!pixels || !ximage)
		return;

	XGCValues gcv;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	GC gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (clip, gc);

	Coordinate x = 0, y = 0;
	Dimension w = image_width, h = image_height;

	bool did_put = false;

	switch (mode) {
	case WIDGET_ALIGN_TOP_LEFT:
		XPutImage (app->dpy, window, gc, ximage, 
			clip.location.x, clip.location.y, clip.location.x, clip.location.y, 
			clip.size.width, clip.size.height);
		did_put = true;
		break;
	case WIDGET_ALIGN_MIDDLE_LEFT:
		y = (rect.size.height - image_height) / 2;
		break;
	case WIDGET_ALIGN_BOTTOM_LEFT:
		y = rect.size.height - image_height;
		break;

	case WIDGET_ALIGN_TOP_CENTER:
		x = (rect.size.width - image_width) / 2;
		break;
	case WIDGET_ALIGN_CENTER: 
		x = (rect.size.width - image_width) / 2;
		y = (rect.size.height - image_height) / 2;
		break;
	case WIDGET_ALIGN_BOTTOM_CENTER:
		x = (rect.size.width - image_width) / 2;
		y = rect.size.height - image_height;
		break;

	case WIDGET_ALIGN_TOP_RIGHT:
		x = rect.size.width - image_width;
		break;
	case WIDGET_ALIGN_MIDDLE_RIGHT:
		x = rect.size.width - image_width;
		y = (rect.size.height - image_height) / 2;
		break;
	case WIDGET_ALIGN_BOTTOM_RIGHT:
		x = rect.size.width - image_width;
		y = rect.size.height - image_height;
		break;

	case WIDGET_ALIGN_STRETCH:
		// Not yet supported.
		break;
	}
	
	if (!did_put) {
		int src_x1 = clip.location.x - x;
		int src_y1 = clip.location.y - y;
		int src_x2 = src_x1 + clip.size.width - 1;
		int src_y2 = src_y1 + clip.size.height - 1;
		int dest_x = clip.location.x;
		int dest_y = clip.location.y;
		int width = clip.size.width;
		int height = clip.size.height;
		if (src_x1 < 0) {
			width += src_x1;
			dest_x -= src_x1;
			src_x1 = 0;
		}
		if (src_y1 < 0) {
			height += src_y1;
			dest_y -= src_y1;
			src_y1 = 0;
		}
		if (src_x2 > w) {
			int diff = w - src_x2;
			width -= diff;
		}
		if (src_y2 > h) {
			int diff = h - src_y2;
			height -= diff;
		}

		if (width > 0 && height > 0)
			XPutImage (app->dpy, window, gc, ximage, src_x1, src_y1, dest_x, dest_y, width, height);
		
#if 0
printf ("src_x1,src_y1=%d,%d\n",src_x1,src_y1);
printf ("dest_x,dest_y=%d,%d\n",dest_x,dest_y);
printf ("width,height=%d,%d\n",width,height);
#endif
		XSetForeground (app->dpy, gc, bg);
		if (x > 0) {
			XFillRectangle (app->dpy, window, gc, 0, 0, x, rect.size.height);
			XFillRectangle (app->dpy, window, gc, rect.size.width-x, 0, x, rect.size.height);
		}
		if (y > 0) {
			XFillRectangle (app->dpy, window, gc, 0, 0, rect.size.width, y);
			XFillRectangle (app->dpy, window, gc, 0, rect.size.height-y, rect.size.width, y);
		}
	}

	XFreeGC (app->dpy, gc);
}

ImageView::ImageView (void) 
{
	delegate = NULL;
	pixels = NULL;
	image_width = 0;
	image_height = 0;
	ximage = NULL;
	mode = WIDGET_ALIGN_CENTER;
}

ImageView::~ImageView ()
{
	if (ximage) {
		XDestroyImage (ximage);
		ximage = NULL;
	}
	if (pixels) {
		free (pixels);
		pixels = NULL;
	}
}

void
ImageView::allocate_pixels ()
{
	if (pixels)
		free (pixels);
	pixels = NULL;

	if (image_width < 1 || image_height < 1) {
		app->log_warning ("Display has invalid dimensions.");
		return;
	}

	unsigned bytes_per_pixel = 3;
	image_ncomponents = 3;
	unsigned size = image_width * image_height * bytes_per_pixel;
	pixels = (char*) calloc (1, size);
	if (!pixels) {
		app->log_warning ("Cannot allocate pixels for image.");
		return;
	}

	if (ximage)
		XDestroyImage (ximage);
		
	ximage = (XImage*) malloc (sizeof(XImage));
	if (!ximage) {
		free (pixels);
		pixels = NULL;
		app->log_warning ("Cannot allocate image.");
		return;
	}

	// Set up an RGB XImage.
	//
	ximage->width = image_width;
	ximage->height = image_height;
	ximage->data = pixels;
	ximage->xoffset = 0;
	ximage->format = ZPixmap;
	ximage->bitmap_unit = 8;
	ximage->bitmap_pad = 8;
	ximage->depth = 24;
	ximage->byte_order = LSBFirst;
	ximage->bitmap_bit_order = LSBFirst;
	ximage->bits_per_pixel = 24;
	ximage->bytes_per_line = 3 * image_width;
	ximage->red_mask = 0xff0000;
	ximage->blue_mask = 0xff00;
	ximage->green_mask = 0xff;
	
	if (!XInitImage (ximage)) {
		app->log_warning ("Unable to init XImage.");
		return;
	}
}

ImageView::ImageView (Size size, RGB image_bg) 
{
	delegate = NULL;
	pixels = NULL;
	image_width = 0;
	image_height = 0;
	ximage = NULL;
	mode = WIDGET_ALIGN_CENTER;

	if (size.width < 1 || size.height < 1)
		return;

	bg = image_bg;

	image_width = (unsigned) size.width;
	image_height = (unsigned) size.height;
	allocate_pixels ();
}

ImageView::ImageView (char *path, RGB image_bg) 
{
	delegate = NULL;
	pixels = NULL;
	image_width = 0;
	image_height = 0;
	ximage = NULL;
	mode = WIDGET_ALIGN_CENTER;

	bg = image_bg;

	if (!path)
		return;
	FILE *f = fopen (path, "rb");
	if (!f)
		return;
	uint8_t b [4];
	int n = fread (b, 1, 4, f);
	fclose (f);
	if (n != 4) {
		app->log_warning ("Bad image file.");
		return;
	}

	//--------------------
	// Identify file type.
	//
	char type = 0;
	if (b[0] == 0xff && b[1] == 0xd8 && b[2] == 0xff)
		type = 'j';

	if (type != 'j') {
		app->log_warning ("Unsupported image file type.");
		return;
	}

	read_jpeg_file (path);
}

/*------------------------------------------------------------------------*/

static int y=0;
void my_error_exit (j_common_ptr cinfo);

struct my_error_mgr {
	struct jpeg_error_mgr pub;
	jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr *my_error_ptr;

char
ImageView::read_jpeg_file (char * filename)
{
	// This method is largely adapted from the jpeglib sample code.
	//
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	int row_stride;
	FILE * infile;
	JSAMPARRAY buffer;

	if (!(infile = fopen (filename, "rb"))) {
		fprintf (stderr, "Can't open JPEG file %s\n", filename);
		return 0;
	}

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;
	if (setjmp(jerr.setjmp_buffer)) {
		jpeg_destroy_decompress (&cinfo);
		fclose (infile);
		return 0;
	}

	jpeg_create_decompress (&cinfo);
	jpeg_stdio_src (&cinfo, infile);
	jpeg_read_header (&cinfo, TRUE);

	cinfo.buffered_image = 0;
	cinfo.dct_method = JDCT_FLOAT;
	cinfo.dither_mode = JDITHER_NONE;
	cinfo.do_block_smoothing = 1;
	cinfo.do_fancy_upsampling = 1;
	cinfo.out_color_space = JCS_RGB;
	cinfo.output_gamma = 1.0;
	cinfo.quantize_colors = 0;
	cinfo.raw_data_out = 0;
	cinfo.scale_denom = 0;
	cinfo.scale_num = 1;
	cinfo.two_pass_quantize = 0;

	jpeg_start_decompress (&cinfo);

	image_width = cinfo.output_width;
	image_height = cinfo.output_height;
	image_ncomponents = cinfo.output_components;
	row_stride = cinfo.output_width * cinfo.output_components;

	allocate_pixels ();

#ifdef DEBUG_IMAGEWIDGET
	printf ("Reading JPEG that is %u x %u, depth %d.\n", (unsigned)image_width, (unsigned)image_height, image_ncomponents);
#endif

	buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	while (cinfo.output_scanline < cinfo.output_height) {
		jpeg_read_scanlines (&cinfo, buffer, 1);
		save_jpeg_scanline (buffer[0], row_stride);
	}

	jpeg_finish_decompress (&cinfo);
	jpeg_destroy_decompress (&cinfo);

	fclose (infile);

	y=0;
	return 1;
}

int 
ImageView::save_jpeg_scanline (JSAMPLE* data, int nbytes)
{
	// This method is largely adapted from the jpeglib sample code.
	//
	uint8_t *tmp = (uint8_t *) data;
	uint8_t *tmp2 = (uint8_t *) (pixels + y * nbytes);

	unsigned bytes_per_pixel = app->dpy_depth / 8;

	int i;
	switch (image_ncomponents){
	case 1: {
		uint8_t byte = tmp [0];
		*tmp2++ = byte;
		*tmp2++ = byte;
		*tmp2++ = byte;
		if (bytes_per_pixel == 4) 
			*tmp2++ = 0;
		} break;
	case 3:
		if (bytes_per_pixel == 3) {
			for (i = 0; i < nbytes; i+=3) {
				tmp2[0] = tmp[2];
				tmp2[1] = tmp[1];
				tmp2[2] = tmp[0];
				tmp += 3;
				tmp2 += 3;
			}
		}
		else if (bytes_per_pixel == 4) {
			for (i = 0; i < nbytes; i+=3) {
				tmp2[0] = tmp[2];
				tmp2[1] = tmp[1];
				tmp2[2] = tmp[0];
				tmp2[3] = 0;
				tmp += 3;
				tmp2 += 4;
			}
		}
		break;
	case 4:
		if (bytes_per_pixel == 3) {
			for (i = 0; i < nbytes; i+=3) {
				tmp2[0] = tmp[2];
				tmp2[1] = tmp[1];
				tmp2[2] = tmp[0];
				tmp += 4;
				tmp2 += 3;
			}
		}
		else if (bytes_per_pixel == 4) {
			for (i = 0; i < nbytes; i+=3) {
				tmp2[0] = tmp[2];
				tmp2[1] = tmp[1];
				tmp2[2] = tmp[0];
				tmp2[3] = tmp[3]; 
				tmp += 4;
				tmp2 += 4;
			}
		}
		break;
	}

	y++;
	return 1;
}

void
my_error_exit (j_common_ptr cinfo)
{
	// This method is largely adapted from the jpeglib sample code.
	//
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	(*cinfo->err->output_message) (cinfo);
	longjmp (myerr->setjmp_buffer, 1);
}

