/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


//#define DEBUG_SCROLLING

#include <stdio.h>
#include <stdlib.h>

#include <TouchWidgets/App.h>
#include <TouchWidgets/Scroll.h>

using namespace TouchWidgets;

Scroll::Scroll () 
{
	delegate = NULL;
	use_momentum = true;
        content_offset = Point_zero ();
        additional_offset_during_touches = Point_zero ();
        allow_horizontal_scrolling = true;
        allow_vertical_scrolling = true;
	scrolling = false;
        children_original_locations = NULL;
	moveable_window = None;
	content_size = Size_new (0, 0);
}

Scroll::~Scroll ()
{
	destroy_window ();

	if (children_original_locations) 
		free (children_original_locations);
}

void
Scroll::destroy_window ()
{
        if (!window)
                return;

        int i;
        for (i = 0; i < n_children; i++) {
                Widget *w = children [i];
                w->destroy_window ();
        }

	if (moveable_window) {
		XDestroyWindow (app->dpy, moveable_window);
                moveable_window = None;
	}
        if (window) {
                XDestroyWindow (app->dpy, window);
                window = None;
        }
}

Window 
Scroll::get_window (void)
{
	if (!window) 
		return None;

	// Provide the moveable window to prospective 
	if (!moveable_window) {
		if (!content_size.width) {
			content_size = rect.size;
		}

		moveable_window = XCreateSimpleWindow (app->dpy, window,
			0, 0, content_size.width, content_size.height, 0, 0, bg);
#ifdef DEBUG_SCROLLING
		printf("Scroll: Created moveable_window, size is %s\n", Size_string (content_size));
#endif
		XSelectInput (app->dpy, moveable_window, ExposureMask);
		XMapWindow (app->dpy, moveable_window);
	}

	return moveable_window;
}

//-----------------------------------------------------------------------------
// Name:	scroll_to
// Purpose:	Provides a way for other classes to cause the scroll container
//		to jump to a specific position.
//-----------------------------------------------------------------------------

void
Scroll::scroll_to (Point p)
{
	if (!window || !moveable_window)
		return;

	content_offset = Point_zero ();
        additional_offset_during_touches = Point_zero ();

	limit_scrolling (&p);
	content_offset = p;

	XMoveWindow (app->dpy, moveable_window, p.x, p.y);
#ifdef DEBUG_SCROLLING
	printf ("Scroll: Moved moveable_window to location %s\n", Point_string (p));
#endif
}

void
Scroll::scroll_start (Point p)
{
	scrolling = true;
        additional_offset_during_touches = p;

	set_backing_store (true);

	XMoveWindow (app->dpy, moveable_window, content_offset.x + p.x, content_offset.y + p.y);

#ifdef DEBUG_SCROLLING
        printf ("Scrolling starts, content_offset = %s\n", Point_string (p));
#endif

}

//-----------------------------------------------------------------------------
// Name:	limit_scrolling
// Purpose:	Private method to prevent motion beyond the content-size area.
//-----------------------------------------------------------------------------

void
Scroll::limit_scrolling (Point *p)
{
	// +---------+
	// +-------+ |
	// |       | |
	// |visible| |
	// | area  | |
	// |       | |
	// +-------+ |
	// | moveable|
	// |  window |
	// +---------+

	// NOTE: The location p is the x,y delta
	// expressing the finger's touch motion,
	// .'. negative x is leftward, and 
	// positive y is downward etc.

	// NOTE: Negative p->x corresponds to 
	// the moveable window's content moving left,
	// i.e. the moveable window's x coordinate
	// decreasing .'. the mininum p->x is 
	// visible's width minus moveable's width.

	// NOTE: Positive p->x means the moveable
	// window's contents move right, its x 
	// increases and .'. the maximum p->x is 0.

	// RULE: If content size is not yet set, prevent all motion.
	//
	if (!content_size.width || !content_size.height) {
		app->log_warning ("Attempt to scroll before content_size is set.");
		*p = Point_zero ();
		return;
	}

	Coordinate minimum_x = rect.size.width - content_size.width;
	Coordinate minimum_y = rect.size.height - content_size.height;
	if (minimum_x > 0)
		minimum_x = 0;
	if (minimum_y > 0)
		minimum_y = 0;

	Coordinate effective_x = p->x + content_offset.x;
	Coordinate effective_y = p->y + content_offset.y;

	// Scrolling leftward
	if (effective_x < minimum_x)
		p->x = minimum_x - content_offset.x;

	// Scrolling upward
	if (effective_y < minimum_y)
		p->y = minimum_y - content_offset.y;

	// Scrolling downward
	if (effective_y > 0)
		p->y = -content_offset.y;

	// Scrolling rightward
	if (effective_x > 0)
		p->x = -content_offset.x;
}

void
Scroll::scroll_continue (Point p)
{
	limit_scrolling (&p);

	float speed = app->get_touch_speed_estimation ();
#ifdef DEBUG_SCROLLING
        printf ("Scrolling continues, speed = %g, delta = %s\n", speed, Point_string (p));
#endif

        additional_offset_during_touches = p;

	XMoveWindow (app->dpy, moveable_window, content_offset.x + p.x, content_offset.y + p.y);

}

void
Scroll::scroll_end (Point p)
{
	float speed = app->get_touch_speed_estimation ();
	limit_scrolling (&p);

#ifdef DEBUG_SCROLLING
        printf ("Scrolling ends, speed was %g, delta = %s\n", speed, Point_string (p));
#endif

	content_offset.x += p.x;
	content_offset.y += p.y;
        additional_offset_during_touches = Point_zero ();

	XMoveWindow (app->dpy, moveable_window, content_offset.x, content_offset.y);

	scrolling = false;

	if (use_momentum) {
		if (speed >= TOUCH_SPEED_THRESHOLD_FOR_MOMENTUM)
			begin_internal_animation (p);
		else
			app->log ("Final touch speed not fast enough to justify momentum.");
	}
	else
		set_backing_store (false);
}

void 
Scroll::begin_internal_animation (Point vector)
{
	if (!use_momentum) 
		return;

	// Continue the motion for .33 seconds.
	// We will cover a distance determined by our
	// current speed.
	//
#define MAX_MOMENTUM_DURATION (10*ANIMATION_SLICES_PER_SECOND)
	internal_animation_counter = MAX_MOMENTUM_DURATION;

	float speed = app->get_touch_speed_estimation ();
	Dimension xdistance = speed * MAX_MOMENTUM_DURATION;
	if (vector.y < 0)
		xdistance = -xdistance;
	Dimension ydistance = speed * MAX_MOMENTUM_DURATION;
	if (vector.y < 0)
		ydistance = -ydistance;

	// To compensate for friction let's increase
	// the distance.
	xdistance *= 1.2;
	ydistance *= 1.2;

	internal_animation_delta_x = xdistance / ANIMATION_SLICES_PER_SECOND;
	internal_animation_delta_y = ydistance / ANIMATION_SLICES_PER_SECOND;

	internal_animation_point = Point_zero ();

	app->begin_internal_animation (this);
}

void 
Scroll::perform_internal_animation (void)
{
	internal_animation_counter--;
	if (internal_animation_counter < 0) {
		end_internal_animation ();
		return;
	}

	float y = content_offset.y+internal_animation_point.y;
#ifdef DEBUG_MOMENTUM
	printf ("internal_animation_counter=%d, y=%g\n",internal_animation_counter, y);
#endif

	internal_animation_point.y += internal_animation_delta_y;
	limit_scrolling (&internal_animation_point);

	XMoveWindow (app->dpy, moveable_window, 
		content_offset.x + internal_animation_point.x, 
		content_offset.y + internal_animation_point.y);

	// Calculate the amount of friction per animation slice.
	float friction = 1 - 0.70 / ANIMATION_SLICES_PER_SECOND;
	internal_animation_delta_y *= friction;

	float top = 0;
	float bottom = rect.size.height - content_size.height;

	if (y >= top && internal_animation_delta_y > 0) {
#ifdef DEBUG_SCROLLING
		puts ("Reached top.");
#endif
		end_internal_animation ();
	}
	else if (y <= bottom && internal_animation_delta_y < 0) {
#ifdef DEBUG_SCROLLING
		puts ("Reached bottom.");
#endif
		end_internal_animation ();
	}

	if (fabs (internal_animation_delta_y) < 3 /* XX */) {
#ifdef DEBUG_SCROLLING
		puts ("Slowed to a stop.");
#endif
		end_internal_animation ();
	}

	// XX To do: implement "bounce".
}

void 
Scroll::end_internal_animation (void)
{
	content_offset.x += internal_animation_point.x;
	content_offset.y += internal_animation_point.y;
	internal_animation_point = Point_zero ();

	app->end_internal_animation (this);
	internal_animation_counter = 0;

	set_backing_store (false);
}

Point 
Scroll::get_content_offset (void)
{
	return Point_new (additional_offset_during_touches.x + content_offset.x,
		 additional_offset_during_touches.y + content_offset.y);
}

void 
Scroll::set_content_size (Size size)
{
	if (size.width < 1 || size.height < 1)
		return;

	content_size = size;

	if (!window || !moveable_window)
		return;

	XResizeWindow (app->dpy, moveable_window, size.width, size.height);
}

void
Scroll::child_was_tapped (int which)
{
	ENTRY
}

