/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <TouchWidgets/App.h>
#include <TouchWidgets/Widget.h>
#include <TouchWidgets/TitleBar.h>
#include <TouchWidgets/ButtonBar.h>
#include <TouchWidgets/Tabs.h>

// #define DEBUG_TAB_CONTROLLER

extern "C" {
#include <TouchWidgets/Util.h>
}

using namespace TouchWidgets;

// Tabs has 3 children:
//	- The title bar.
//	- The content that is provided by a sub-controller.
// 	- The button bar.

Tabs::Tabs () 
{
	delegate = NULL;
	selection = -1;
	current_content_root = NULL;

	use_animation = true;
	animation_duration = DEFAULT_SLIDE_TIME;
	animations_count = 0;

	// RULE: Create the widgets as early as possible, 
	// so that the Tabs is "ready for business"
	// before its root widget has been attached
	// to a parent widget.

	root_widget = new Widget ();
	root_widget->bg = RGB_WHITE;
	root_widget->name = "Tabs TOP LEVEL root_widget";

	title_bar = new TitleBar ((char*) "Title");
	title_bar->name = "Tabs title_bar";
	button_bar = new ButtonBar ();
	button_bar->name = "Tabs button_bar";
	button_bar->set_delegate (this);

	root_widget->add_child (title_bar);
	root_widget->add_child (button_bar);
}

Tabs::~Tabs () 
{
	if (current_content_root) 
		deselect ();

        if (children) {
#ifdef DEBUG_TAB_CONTROLLER
		printf ("TAB has %d child controllers to remove.\n", n_children);
#endif
		// By this point, all child controllers' root widgets should be disconnected.
		// If not the default destructor will take care of them.
		// 
                for (int i = n_children-1; i >= 0; i--) {
                        delete children[i];
		}
                free (children);
                children = NULL;
        }

        if (parent) {
                parent->remove_child (this);
                parent = NULL;
        }

	// By this point, our root widget and its two child widgets
	// are isolated.
	//
	delete root_widget;
	root_widget = NULL;
}

void 
Tabs::destroy_widgets (void)
{
	ENTRY

	if (root_widget) {
		delete root_widget;
		root_widget = NULL;
	}
}

void 
Tabs::layout_widgets (void)
{
	ENTRY

	if (!root_widget)
		return;

	Rect r = root_widget->get_rect ();
	int w = r.size.width;
	int h = r.size.height;
	title_bar->set_rect (Rect_new (0, 0, w, 44));
	button_bar->set_rect (Rect_new (0, h-44, w, 44));

	if (current_content_root) {
		current_content_root->set_rect (Rect_new (0, 44, w, h-88));
		Controller *child = children [selection];
		child->layout_widgets ();
		current_content_root->draw (current_content_root->get_bounds ());
	}
}

void
Tabs::will_appear (void)
{
}

void
Tabs::did_appear (void)
{
	if (n_children) {
		select_nth_instantaneous (0);
		button_bar->select (0);
	}
}

void 
Tabs::will_disappear (void)
{
	ENTRY

}

void 
Tabs::deselect (void)
{
	if (animations_count > 0)
		return;

#ifdef DEBUG_TAB_CONTROLLER
	printf("Tabs: did deselect, current selection = %d\n", selection);
#endif
	if (!root_widget) 
		return;
	if (selection < 0 || selection >= n_children)
		return;

	Controller *previous_child = children [selection];
	selection = -1;
	previous_child->will_disappear ();
	current_content_root->set_hidden (true);
	current_content_root->remove_from_parent ();
	current_content_root = NULL;
}

void 
Tabs::select_nth (int which)
{
	if (animations_count > 0)
		return;

	if (selection >= 0 && which == selection)
		return;

	if (!use_animation || animation_duration < 2*LIMIT_OF_HUMAN_PERCEPTION)
		select_nth_instantaneous (which);
	else
		select_nth_animated (which);
}

void
Tabs::select_nth_instantaneous (int which)
{
#ifdef DEBUG_TAB_CONTROLLER
	printf("Tabs: selected %d, instantaneous\n", which);
#endif
	if (!root_widget) 
		return;
	if (!n_children)
		return;
	if (selection >= 0) 
		deselect ();
	if (which < 0 || which >= n_children)
		return;

	selection = which;

	Controller *child = children [selection];
	current_content_root = child->get_root_widget ();

	if (current_content_root) {
		child->will_appear ();
		root_widget->add_child (current_content_root);
		layout_widgets ();
		child->layout_widgets ();
		child->create_windows ();
		current_content_root->set_hidden (false);
	} else {
		app->log_warning ("Child controller of tab controller has no widgets.");
	}
}

void 
Tabs::select_nth_animated (int which)
{
#ifdef DEBUG_TAB_CONTROLLER
	printf("Tabs: selected %d, animated\n", which);
#endif

	if (!root_widget) 
		return;
	if (!n_children)
		return;
	if (which < 0 || which >= n_children)
		return;

	bool leftward = false;
	if (which < 0 || which > selection)
		leftward = true;

	//------------------------------------------------------------
	// Start two animations: 
	// 1. Current controller's root widget moves off-screen.
	// 2. Incoming controller's root widget moves off-screen.
	//
        Rect r = root_widget->get_rect ();
        int w = r.size.width;
        int h = r.size.height;

	//----------
	// Outgoing.
	//
	if (selection >= 0) {
        	Rect rr = current_content_root->get_rect ();
		
		if (leftward)
			rr.location.x -= w;
		else
			rr.location.x += w;

		current_content_root->set_animation_delegate (this);
		current_content_root->animate_to_rect (rr, animation_duration);

		outgoing_content_root = current_content_root;
		outgoing_controller = selection;
		animations_count++;

		// RULE: While animation is happening these are not valid, because
		// the animation must run to completion before anything can happen.
		//
		current_content_root = NULL;
		selection = -1;
	}

	//----------
	// Incoming.
	//
	if (which >= 0) {
		Controller *child = children [which];
		//child->will_appear ();

		Widget *rw = child->get_root_widget ();
		rw->set_hidden (false);
		root_widget->add_child (rw);
		Rect rect_for_incoming = Rect_new (leftward ? w : -w, 44, w, h-88); // From layout_widgets.
		rw->set_rect (rect_for_incoming);
		
		child->layout_widgets ();
		child->create_windows ();

#ifdef DEBUG_TAB_CONTROLLER
printf ("STARTING RECT %s, %x\n", Rect_string (rect_for_incoming), (int) rw->get_window ());
#endif
		rect_for_incoming.location.x = 0;
#ifdef DEBUG_TAB_CONTROLLER
printf ("ENDING RECT %s, %x\n", Rect_string (rect_for_incoming), (int) rw->get_window());
#endif
		rw->set_hidden (false);
		rect_for_incoming = Rect_new (0, 44, w, h-88); // From layout_widgets.
		rw->set_animation_delegate (this);
		rw->animate_to_rect (rect_for_incoming, animation_duration);

		incoming_controller = which;
		incoming_content_root = rw;
		animations_count++;
	} 
}

void 
Tabs::add_button_and_child (const char *str, Controller *controller)
{
	if (!controller) {
		app->log_warning ("Null controller given to Tabs::add_button_and_child.");
		return;
	}

	if (!str)
		str = app->translate ("Untitled");

	button_bar->add_button (str);
	add_child (controller);
}

void 
Tabs::remove_button_and_child (int)
{
	// Not yet implemented.
}

void
Tabs::receive_message (Object *sender, Message msg, MessageParameter tmp)
{
	if (sender == button_bar) {
		int which = (int) tmp;

		if (!strcmp (msg, BUTTONBAR_SELECTION)) {
#ifdef DEBUG_TAB_CONTROLLER
			printf ("Tabs told tab %d selected.\n", which);
#endif
			select_nth (which);
		}
	}
	else if (sender == title_bar) {
		// Future buttons on title bar.
	}
	else if (sender == outgoing_content_root) {
		if (outgoing_controller < 0 || outgoing_controller >= n_children)
			return;
		Controller *child = children [outgoing_controller];
		child->will_disappear ();

#if 0
		Widget *root = child->get_root_widget ();
		root->destroy_window ();
#else
		child->destroy_windows ();
#endif

		animations_count--;
#ifdef DEBUG_TAB_CONTROLLER
		puts ("OUTGOING CONTROLLER ANIMATION COMPLETED");
#endif
	}
	else if (sender == incoming_content_root) {
		if (incoming_controller < 0 || incoming_controller >= n_children)
			return;
		Controller *child = children [incoming_controller];
#ifdef DEBUG_TAB_CONTROLLER
		printf ("INCOMING CONTROLLER ANIMATION COMPLETED, root_widget=0x%08lx\n", (unsigned long) incoming_content_root);
#endif
		current_content_root = incoming_content_root;
		current_content_root->set_hidden (false);
		current_content_root->set_backing_store (true);
		selection = incoming_controller;
		animations_count--;
	}

#ifdef DEBUG_TAB_CONTROLLER
	printf ("animations_count = %d\n", animations_count);
#endif
}

void 
Tabs::set_title (const char* s)
{
	if (s && root_widget && title_bar) {
		title_bar->set_title (s);
	}
}

