/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


#include <stdio.h>
#include <wchar.h>

#include <TouchWidgets/App.h>
#include <TouchWidgets/TextField.h>

using namespace TouchWidgets;

TextField::TextField () 
{
	delegate = NULL;
	cursor_position = 0;
	length = 0;
	full = false;
	have_focus = false;
	ctrl = false;
	shift = false;

	rect = Rect_zero ();
	
	fg = app->fg;
	bg = app->bg;
	highlight_color = app->highlight_color;
	// highlight_bgcolor = fg_;

	border = TEXTENTRY_BORDER;

	erase ();
}

TextField::~TextField ()
{
}

void 
TextField::receive_message (Object *sender, Message msg, MessageParameter param)
{
	if (!strcmp (msg, KEY_DOWN)) {
		key_press (param);
	}
	else
	if (!strcmp (msg, KEY_UP)) {
		key_release (param);
	}
	else
	if (!strcmp (msg, TOUCH_DOWN_INSIDE)) {
		if (!have_focus) {
			cursor_position = length;
			have_focus = true;
			app->request_key_focus (this);
			redraw ();
		} else {
			// If the user has clicked in the widget during
			// data entry, it's to move the cursor.
			// So let's locate where the cursor should go.
			//
			Coordinate x = TEXTENTRY_BORDER + TEXTENTRY_PADDING;
			if (length) {
				int i=0;
				while (x >= (char_positions[i] + char_positions[i+1])/2) {
					i++;
				}
				cursor_position = i;
			}
			else
				cursor_position = 0;

			if (cursor_position > length)
				cursor_position = length;
		}
	}
}

void 
TextField::focus_taken ()
{
	have_focus = false;

	redraw ();
}

//--------------------------------------------------------------------
// Name:	erase
// Purpose:	Erase the line.
//--------------------------------------------------------------------
void 
TextField::erase ()
{
	length = 0;
	cursor_position = 0;
	line[0] = 0;

	redraw ();
}

//--------------------------------------------------------------------
// Name:	add_integer
// Purpose:	Add an integer to the line.
//--------------------------------------------------------------------
void 
TextField::add_integer (int n)
{
	char str[30];
	snprintf (str, 29, "%d", n);
	add_string (str);
}

//--------------------------------------------------------------------
// Name:	add_char
// Purpose:	Add a character to the line.
//--------------------------------------------------------------------
void 
TextField::add_char (char ch)
{	
	char str[2] = {ch, 0};
	add_string (str);
}

//--------------------------------------------------------------------
// Name:	get_text
// Purpose:	Retrieves pointer to internal buffer.
//--------------------------------------------------------------------
char*
TextField::get_text (void)
{
	return line;
}

#if 0
//--------------------------------------------------------------------
// Name:	get_text
// Purpose:	Retrieves wide form of text.
//--------------------------------------------------------------------
int 
TextField::get_text (wchar_t *str, int maxlen)
{
	if (!str || maxlen < 1)
		return 0;

	int i = 0;
	while (i < length && i < (maxlen-1)) {
		str[i] = (char) line[i];
		i++;
	}
	str[i] = 0;
	return i;
}
#endif

//--------------------------------------------------------------------
// Name:	add_string
// Purpose:	Add a string to the line.
//--------------------------------------------------------------------
void 
TextField::add_string (char *str)
{
	if (!str)
		return;

	while (*str) {
		char ch = *str++;

		if (length >= TEXTENTRY_MAXLEN) {
			return;
		}

		if (ch < ' ' || ch > 126) {
			return;
		}

		if (cursor_position == length) {
			line [length++] = ch;
			if (length < TEXTENTRY_MAXLEN)
				line [length] = 0;
			cursor_position = length;
		} 
		else if (!full && length < TEXTENTRY_MAXLEN-1) {
			int i = length;
			while (i >= 0 && i >= cursor_position) {
				line[i] = line[i-1];
				i--;
			}
			line [cursor_position++] = ch;
			length++;
			if (length < TEXTENTRY_MAXLEN)
				line [length] = 0;
		}
	}

	redraw ();
}

//--------------------------------------------------------------------
// Name:	draw
// Purpose:	Paint the widget.
// Note:	Called with GUI lock held.
//--------------------------------------------------------------------
void 
TextField::draw (Rect clip)
{
	if (!window || hidden)
		return;

	char tmp[2];
	int x, y, x2;
	int cursor_height;
	int cursor_y;
	RGB cursor_color;

        XGCValues gcv;
        memset ((void*) &gcv, 0, sizeof(XGCValues));
        GC gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (clip, gc);

        XSetFont (app->dpy, gc, app->default_font->fid);

        Dimension w = rect.size.width;
        Dimension h = rect.size.height;

        XSetForeground (app->dpy, gc, bg);
        XFillRectangle (app->dpy, window, gc, border, border, w - 2*border, h - 2*border);

	draw_3d_border (clip, false);

        Dimension ascent = app->default_font->ascent;
        Dimension th = ascent + app->default_font->descent;

	int width = rect.size.width;
	int height = rect.size.height;
	cursor_height = th + 4;
	cursor_y = (height - cursor_height)/2;
	cursor_color = RGB_AMBER;
	x = TEXTENTRY_BORDER + TEXTENTRY_PADDING;
	//y = (height - th)/2;
	y = ascent + border + ((height-2*border) - ascent)/2,

	tmp[1] = 0;

	int i;
	full = false;

	//------------------------------------------------------------
	// Draw each character individually so that we can record 
	// individual character positions, limit length and control 
	// spacing.
	//
	int cursor_x = -1;
	for (i = 0; i < length && !full; i++) {
		tmp[0] = line[i];
		char_positions[i] = x;

        	XSetForeground (app->dpy, gc, fg);
		XDrawString (app->dpy, window, gc, x, y, tmp, 1);

        	Dimension tw = XTextWidth (app->default_font, tmp, 1);
		x2 = x + tw;
		
		// Draw the cursor.
		//
		if (i == cursor_position) {
			cursor_x = x;
		}
		x = x2 + 1;
		
		if (x > (width - 6))
			full = true;
		else 
			full = false;
	}
	char_positions[i] = x;
	if (cursor_x < 0)
		cursor_x = x;

	if (have_focus) {
		XSetForeground (app->dpy, gc, cursor_color);
		XDrawLine (app->dpy, window, gc, cursor_x, cursor_y, cursor_x, cursor_y + cursor_height - 1);
		XDrawLine (app->dpy, window, gc, cursor_x+1, cursor_y, cursor_x+1, cursor_y + cursor_height - 1);
	}
}

//--------------------------------------------------------------------
// Name:	key_release
// Purpose:	Key release handler.
//--------------------------------------------------------------------

void 
TextField::key_release (KeySym key)
{
	if (!window || hidden)
		return;

	switch (key) {
	case XK_Shift_L:
	case XK_Shift_R:
		shift = false;
		break;
	case XK_Control_L:
	case XK_Control_R:
		ctrl = false;
		break;
	}
}

//--------------------------------------------------------------------
// Name:	key_press
// Purpose:	Key press handler.
//--------------------------------------------------------------------

void 
TextField::key_press (KeySym key)
{
	int i;

	if (!window || hidden)
		return;

	//--------------------
	// Perform shifting.
	//
	if (shift) {
		key = get_shifted_keysym (key);
	}
	else if (ctrl) {
		if (key >= XK_a && key <= XK_z) 
			key = key - XK_a + 1;
	}

	//--------------------
	// Process keys.
	//
	switch ((unsigned) key) {
	case XK_Return:
		app->release_key_focus (this);
		have_focus = false;

		// Report that it's finished.
		if (delegate)
			send_message (delegate, DONE_EDITING);

		app->dismiss_keyboard ();
		break;

	case XK_Shift_L:
	case XK_Shift_R:
		shift = true;
		return;

	case XK_Control_L:
	case XK_Control_R:
		ctrl = true;
		return;

	case XK_Escape:
		erase ();
		break;

	case XK_Up:
	case XK_Down:
		// XX Access history
		return;

	case 22: // CTRL-V
		// XX Paste
		return;

	case 6: // CTRL-F
	case XK_Right:
		if (cursor_position < length)
			cursor_position++;
		break;

	case 2: // CTRL-B
	case XK_Left:
		if (cursor_position > 0)
			cursor_position--;
		break;

	case XK_Tab:
		break;

	case XK_BackSpace:
		if (!length)
			return;
		if (cursor_position >= length) {
			line[--length] = 0;
			cursor_position = length;
		} 
		else if (cursor_position > 0) {
			cursor_position--;
			for (i = cursor_position; i < length-1; i++)
				line[i] = line[i+1];
			line[--length] = 0;
		}
		break;

	case 1:	// CTRL-A = Move to start.
		cursor_position = 0;
		break;
	case 5:	// CTRL-E = Move to end.
		cursor_position = length;
		break;
	case 11: // CTRL-K = Clear to end.
		length = cursor_position;
		line [length] = 0;
		break;
	case 4: // CTRL-D = Delete char to right of cursor.
		if (cursor_position < length) {
			int i = cursor_position;
			while (i < length-1) {
				line [i] = line [i+1];
				i++;
			}
			--length;
			line [length] = 0;
		} 
		break;
	
	default:
		if (key >= 0x20 && key < 0x7f) {
			if (length < TEXTENTRY_MAXLEN-1) {
				add_char (key);
			} else {
				app->beep (440 /* Hz */, 250 /* milliseconds */);
			}
		}
		else
			printf ("Unknown keysym: 0x%04x\n", (unsigned) key);
		break;
	}

	redraw ();
}

