/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/


// #define DEBUG_CREATION

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <typeinfo>

#include <TouchWidgets/App.h>
#include <TouchWidgets/Widget.h>
#include <TouchWidgets/Container.h>
#include <TouchWidgets/Scroll.h>

using namespace TouchWidgets;

#define DEFAULT_MAX_CHILDREN (16)

#ifdef DEBUG_CREATION
static int tw_counter = 0;
static int tw_counter2 = 0;
static int w_counter = 0;
static int w_counter2 = 0;
#endif

void 
Widget::create_window (Window parent_window)
{
	if (window) {
		app->log_warning ("create_window: Widget already has a window.");
		return;
	}
	if (!parent_window) {
		app->log_warning ("create_window: No parent window provided.");
		return;
	}

	if (rect.size.width < 1 || rect.size.height < 1) {
		app->log_warning ("Widget has zero dimensions, cannot create its window.");
#ifdef DEBUG_CREATION
		printf ("Widget %s has zero dimensions, cannot create its window.\n", name);
#endif
		return;
	}

	if (automatic_styling) 
		update_styles ();
	else
		printf("automatic_styling==false for %s\n", name);

        window = XCreateSimpleWindow (app->dpy, parent_window, 
		rect.location.x, 
		rect.location.y, 
		rect.size.width, 
		rect.size.height, 0, 0, bg);
        XSelectInput (app->dpy, window, ExposureMask);

	if (!hidden)
		XMapWindow (app->dpy, window);

#ifdef DEBUG_CREATION
	w_counter++;
	printf ("# Windows created = %d, destroyed = %d\n", w_counter, w_counter2);
	if (!hidden)
		printf ("Created & mapped window (id = %08lx) for widget %s at %s.\n", (unsigned long)window, name, Rect_string (rect));
	else
		printf ("Created & DID NOT MAP window (id = %08lx) for widget %s at %s.\n", (unsigned long)window, name, Rect_string (rect));
#endif
}

Rect
Widget::get_bounds ()
{
	return Rect_new (0, 0, rect.size.width, rect.size.height);
}

//-----------------------------------------------------------------------------
// Name:	get_rect
// Purpose:	Determine relative rectangle of a widget, with the 
//		qualification that for widgets inside scroll containers we
//		need to add the scroll offset.
//-----------------------------------------------------------------------------

Rect 
Widget::get_rect () 
{
	// We need to adjust the Rect for any Scrolls
	// that this widget may be contained in. This can be
	// done by traversing up to the root.
	//
	Rect r = rect;
	Widget *par = parent;
	while (par) {
		if (dynamic_cast<Scroll*>(par)) {
			Scroll *scroll = (Scroll*) par;
			Point offset = scroll->get_content_offset ();
			r.location.x += offset.x;
			r.location.y += offset.y;
#ifdef DEBUG_SCROLLING
			printf ("get_rect: Adding scroll container offset to reported widget rect.\n");
#endif
		}
		par = par->parent;
	}

	return r;
}

void 
Widget::message (Message* m)
{
        // RULE: Basic widget does nothing with incoming message.
	// RULE: It is the sender that deallocates messages, not the recipient. 
}

#if 0
void
Widget::reparent_window (Window new_parent)
{
	if (!window || !new_parent)
		return;

	XReparentWindow (app->dpy, window, new_parent, 
		rect.location.x, 
		rect.location.y);
} 
#endif

Window 
Widget::get_window (void)
{
	return window;
}

void
Widget::set_rect (Rect r)
{
	if (!Rect_equal (r, rect)) {
		rect = r;

		if (window) {
			XMoveResizeWindow (app->dpy, window, rect.location.x, rect.location.y, 
				rect.size.width, 
				rect.size.height);
		}
	}
}

void
Widget::move (Point p)
{
	rect.location = p;

	if (window) {
		XMoveResizeWindow (app->dpy, window, rect.location.x, rect.location.y, 
			rect.size.width, 
			rect.size.height);
	}
}

void
Widget::clear ()
{
#if 0
	XGCValues gcv;
	GC gc;

	memset ((void*) &gcv, 0, sizeof(XGCValues));
	gc = XCreateGC (app->dpy, window, 0, &gcv);

	XSetForeground (app->dpy, gc, bg);
	XFillRectangle (app->dpy, window, gc, 0, 0, rect.size.width, rect.size.height);

	XFreeGC (app->dpy, gc);
#endif
}

void 
Widget::clip_into (Rect r, GC gc) 
{
	if (!gc || !window) return;
	XRectangle rr;
	rr.x = r.location.x;
	rr.y = r.location.y;
	if (!(rr.width = r.size.width))
		return;
	if (!(rr.height = r.size.height))
		return;
	XSetClipRectangles (app->dpy, gc, 0, 0, &rr, 1, Unsorted);
}

void 
Widget::animation_completed (void) 
{
	animating = false;
	animation_origin = Rect_zero ();
	animation_destination = Rect_zero ();
	animation_duration = 0;
	animation_counter = 0;

	if (animation_delegate) 
		send_message (animation_delegate, FINISHED_ANIMATING);
}

bool
Widget::inside (int x_, int y_)
{
	int x1, y1;

	x1 = rect.location.x + rect.size.width - 1;
	y1 = rect.location.y + rect.size.height - 1;
	return x_ >= rect.location.x && x_ <= x1 && y_ >= rect.location.y && y_ <= y1;
}

bool
Widget::inside (Point p)
{
	int x1, y1;

	x1 = rect.location.x + rect.size.width - 1;
	y1 = rect.location.y + rect.size.height - 1;
	return p.x >= rect.location.x && p.x <= x1 && p.y >= rect.location.y && p.y <= y1;
}

void 
Widget::reallocate_children (int n)
{
	if (n < 4)
		n = 4;
	if (n <= max_children)
		n = 2*max_children;

	int new_max_children = n;
	children = (Widget**) realloc (children, sizeof(Widget*) * new_max_children);

	if (!children)
		app->log_fatal ("Out of memory.");

	for (unsigned i = n_children; i < new_max_children; i++)
		children [i] = NULL;

	max_children = new_max_children;
}

//-----------------------------------------------------------------------------
// Name:	add_child
// Purpose:	Performs the narrow task of adding a child widget, but only
//		if the child is not already in the children array.
//-----------------------------------------------------------------------------

void 
Widget::add_child (Widget* w)
{
	if (!w) {
		app->log_fatal ("Attempt to add NULL widget pointer to widget tree.");
	}

	if (!children)
		reallocate_children (DEFAULT_MAX_CHILDREN);

	// Check whether widget is already a child.
	//
	int i;
	for (i = 0; i < n_children; i++) {
		if (w == children [i]) {
			return;
		}
	}

	// RULE: If children array is full, increase
	// its size by 50%.
	//
	if (n_children && n_children >= max_children)
		reallocate_children (max_children + max_children/2);

	children [n_children++] = w;
	w->parent = this;
}

//-----------------------------------------------------------------------------
// Name:	remove_child
// Purpose:	Performs the narrow task of removing a child widget.
//-----------------------------------------------------------------------------

void 
Widget::remove_child (Widget *w)
{
	if (!children || !n_children || !w)
		return;

	int ix = -1;
	int i = 0;
	while (i < n_children) {
		if (children [i] == w) {
			ix = i;
			break;
		}
		i++;
	}
	if (ix < 0) {
		app->log_warning ("Attempt to remove child widget that was not found.");
		return;
	}

	int limit = n_children-1;
	while (ix < limit) {
		children[ix] = children[ix+1];
		ix++;
	}
	n_children--;
}

void 
Widget::remove_from_parent (void)
{
	if (parent) {
		parent->remove_child (this);
		parent = NULL;
		destroy_window ();
	}
}

Widget* 
Widget::get_nth_child (int n)
{
	if (!children || n < 0 || n >= n_children)
		return NULL;
	//==========

	return children [n];
}

void
Widget::set_backing_store (bool on)
{
	if (!window)
		return;

	XSetWindowAttributes attr;
	bzero (&attr, sizeof(XSetWindowAttributes));

	if (on) 
		attr.backing_store = WhenMapped;
	else 
		attr.backing_store = NotUseful;
	
	XChangeWindowAttributes (app->dpy, window, CWBackingStore, &attr);

	int i;
	for (i = 0; i < n_children; i++) {
		Widget *w = get_nth_child (i);
		w->set_backing_store (on);
	}
}

void 
Widget::insert_child_at (Widget* w, int ix)
{
	if (!children)
		reallocate_children (DEFAULT_MAX_CHILDREN);

	if (ix >= n_children) {
		add_child (w);
		return;
	}

	// XX not finished
}

void
Widget::draw (Rect r)
{
        if (!window || hidden)
                return;
}

void 
Widget::draw_3d_border (Rect clip, bool outward)
{
	if (!window)
		return;

	XGCValues gcv;
	GC gc;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (clip, gc);

        Dimension w = rect.size.width;
        Dimension h = rect.size.height;

        RGB dark = RGB_change_intensity (bg, -.2); 
        RGB medium2 = RGB_change_intensity (bg, -.1); 
        RGB medium = RGB_change_intensity (bg, -.5); 
        RGB light2 = RGB_change_intensity (bg, .3); 
        RGB light = RGB_change_intensity (bg, .6); 

	XSetForeground (app->dpy, gc, bg);
	XFillRectangle (app->dpy, window, gc, border, border, w-2*border, h-2*border);

	int i;
	for (i=0; i<border; i++)
	{
		XSetForeground (app->dpy, gc, outward ? light : dark);
		XDrawLine (app->dpy, window, gc, i, i, i, h-1-i);
		XDrawLine (app->dpy, window, gc, i, i, w-1-i, i);

		XSetForeground (app->dpy, gc, !outward ? light2 : dark);
		XDrawLine (app->dpy, window, gc, w-1-i, 1+i, w-1-i, h-1-i);
		XDrawLine (app->dpy, window, gc, i, h-1-i, w-1-i, h-1-i);
	}
	XFreeGC (app->dpy, gc);
}

void 
Widget::draw_gradient_background (Rect clip)
{
	draw_gradient_in_rect (clip, get_bounds ());
}

void 
Widget::draw_gradient_in_rect (Rect clip, Rect rr)
{
	if (!window || !rr.size.height || !rr.size.width)
		return;

	if (bg == bg_gradient) {
		// clear ();	
		return;
	}

	float red_start = (bg >> 16) & 255;
	float green_start = (bg >> 8) & 255;
	float blue_start = bg & 255;

	float red_end = (bg_gradient >> 16) & 255;
	float green_end = (bg_gradient >> 8) & 255;
	float blue_end = bg_gradient & 255;

	XGCValues gcv;
	GC gc;
	memset ((void*) &gcv, 0, sizeof(XGCValues));
	gc = XCreateGC (app->dpy, window, 0, &gcv);

	clip_into (clip, gc);

	int i;
	int w = (int) rr.size.width; 
	int h = (int) rr.size.height; 
	Coordinate x = rr.location.x;
	Coordinate y = rr.location.y;
	for (i = 0; i < h; i++) {
		float tmp = i / (float) h;
		float red = red_start + (red_end-red_start) * tmp;
		float green = green_start + (green_end-green_start) * tmp;
		float blue = blue_start + (blue_end-blue_start) * tmp;
		int r = (int) red;
		int g = (int) green;
		int b = (int) blue;
		if (r > 255)
			r = 255;
		if (g > 255)
			g = 255;
		if (b > 255)
			b = 255;
		int rgb = (r << 16) | (g << 8) | b;

		XSetForeground (app->dpy, gc, rgb);
		XDrawLine (app->dpy, window, gc, x, y + i, x + w - 1, y + i);
	}

	XFreeGC (app->dpy, gc);
}

Widget::Widget (void) 
{
	delegate = NULL;
        bg = RGB_WHITE; 
	fg = RGB_BLACK; 
	bg_gradient = RGB_BLACK; 
	highlight_color = RGB_BLACK; 
	selected_color = RGB_BLACK; 
	border_color = RGB_BLACK;
        border = 0;
        rect = Rect_zero();
        animating = false;
        animation_origin = Rect_zero();
        animation_destination = Rect_zero();
        animation_duration = 0;
	animation_counter = 0;
	animation_delegate = NULL;
	window = 0;
	parent = NULL;
	next = NULL;
	n_children = 0;
	max_children = 0;
	children = NULL;
	automatic_styling = true;
	hidden = false;

#ifdef DEBUG_CREATION
	tw_counter++;
	printf ("# TouchWidgets created = %d, destroyed = %d\n", tw_counter, tw_counter2);
#endif

	rect = Rect_zero ();

        fg = app->fg; 
        bg = app->bg; 
	bg_gradient = app->bg_gradient;
        highlight_color = app->highlight_color; 
        selected_color = app->selected_color;
        border_color = app->border_color;
}

Widget::~Widget ()
{
	if (children) {
		// Remove them in reverse order so that, if anyone calls
		// remove_from_parent, it won't result in a double free.
		//
		for (int i = n_children-1; i >= 0; i--) {
			delete children[i];
		}
		free (children);
		children = NULL;
	}

#ifdef DEBUG_CREATION
	tw_counter2++;
	printf ("# TouchWidgets created = %d, destroyed = %d\n", tw_counter, tw_counter2);
#endif

	parent = NULL;
}

void
Widget::set_hidden (bool hidden_)
{
	hidden = hidden_;
	if (!window) {
		return;
	}
	if (hidden)
		XUnmapWindow (app->dpy, window);
	else
		XMapWindow (app->dpy, window);
}

// Class method.
Point 
Widget::convert_absolute_point_to_relative (Point absolute_point, Widget* w)
{
	Point p = absolute_point;
	while (w) {
		Rect r = w->get_rect ();
		p.x -= r.location.x;
		p.y -= r.location.y;
		w = w->parent;
	}
	return p;
}

void 
Widget::animate_to_rect (Rect r, float duration)
{
	if (animating) {
		app->log_warning ("Cannot start animation because we are already animating.");
		return;
	}

	//------------------------------
	// Convert duration to multiples 
	// of the preset animation slices.
	//
	animation_duration = (unsigned) floorf (duration * ANIMATION_SLICES_PER_SECOND + .5);
	animation_counter = 0;
	animation_origin = rect;
	animation_destination = r;
	animating = true;
	app->add_animation (this);
}

void 
Widget::generate_next_animation_rect (void)
{
	if (!animating)
		return;

	++animation_counter;

	float tmp = ((float) animation_counter) / (float) animation_duration;
	float dx = animation_destination.location.x - animation_origin.location.x;
	float dy = animation_destination.location.y - animation_origin.location.y;
	float dw = animation_destination.size.width - animation_origin.size.width;
	float dh = animation_destination.size.height - animation_origin.size.height;
	float x = animation_origin.location.x + tmp * dx;
	float y = animation_origin.location.y + tmp * dy;
	float w = animation_origin.size.width + tmp * dw;
	float h = animation_origin.size.height + tmp * dh;
	Rect r = Rect_new (x, y, w, h);

	if (tmp <= 1.) 
		set_rect (r);
	else
		animation_completed ();
}

static Widget *
identify_child_at_point_inner (Widget *w, Point *p /* in/out */, bool ignore_container)
{
	if (!w || w->get_hidden ())
		return NULL;

	Rect r = w->get_rect ();
	Point p2 = *p;
	p2.x -= r.location.x;
	p2.y -= r.location.y;

	// If point is outside of the bounds of this widget,
	// we're done.
	//
	if (p2.x < 0 || p2.y < 0 || p2.x >= r.size.width || p2.y >= r.size.height)
		return NULL;

	// Inside of rect, but no children means automatic match.
	if (!w->n_children) {
		*p = p2;
		return w;
	}

	// RULE: If the touched widget is a container,
	// we stop at the container because it has to
	// differentiate between kinds of gestures.
	//
	// Stop at a container -- XX unless the container specifies to pass thru touches.
	//
	if (!ignore_container && dynamic_cast<Container*>(w)) {
		*p = p2;
		return w;
	}

	for (int i = 0; i < w->n_children; i++) {
		Widget *w2 = w->get_nth_child (i); 
		Point p3 = p2;
		Widget *w3 = identify_child_at_point_inner (w2, &p3, ignore_container);
		if (w3) {
			*p = p3;
			return w3;
		}
	}

	*p = p2;
	return w;
}

//-----------------------------------------------------------------------------
// Name:	identify_child_at_relative_point
// Purpose:	Locates a child widget given a point in this widget's parent's 
//		coordinate system. If the first widget found is a container,
//		it does not stop there.
//-----------------------------------------------------------------------------

Widget*
Widget::identify_child_at_relative_point (Point p, Point *return_relative_point)
{
	p.x += rect.location.x;
	p.y += rect.location.y;

	if (return_relative_point)
		*return_relative_point = p;

	Widget *w = identify_child_at_point_inner (this, return_relative_point, true);
	return w;
}

//-----------------------------------------------------------------------------
// Name:	identify_child_for_absolute_point
// Purpose:	Locates a child widget given an absolute point.
//		If a container is encountered, it stops.
//-----------------------------------------------------------------------------

Widget*
Widget::identify_child_for_absolute_point (Point p, Point *return_relative_point)
{
	if (return_relative_point)
		*return_relative_point = p;
	Widget *w = identify_child_at_point_inner (this, return_relative_point, false);
	return w;
}

static void create_windows_recursive (Widget *w, Window parent_window)
{
	if (!w) 
		return;

	w->create_window (parent_window);

	if (!w->n_children)
		return;

	for (int i = 0; i < w->n_children; i++) {
		Widget *child = w->get_nth_child (i);
		create_windows_recursive (child, w->get_window ());
	}
}

void
Widget::create_windows_for_self_and_children (Window parent_window)
{
	if (!parent_window) {
		app->log_warning ("create_windows_for_self_and_children: No parent window provided.");
		return;
	}

	create_windows_recursive (this, parent_window);
}

void
Widget::destroy_window (void)
{
	for (int i = 0; i < n_children; i++) {
		Widget *child = get_nth_child (i);
		child->destroy_window ();
	}

	if (window) {
#ifdef DEBUG_CREATION
	w_counter2++;
	printf ("# Windows created = %d, destroyed = %d\n", w_counter, w_counter2);
	printf ("Destroying window of %s. It was %x\n", name, (int) window);
#endif
		XDestroyWindow (app->dpy, window);
		window = None;
	}
}

void
Widget::update_styles (void)
{
	char tmp [256];
	RGB color;

	const char *cpp_class = class_name ();

	//--------------------
	// Background color.
	//
	if (true) {
		const char *s = app->fetch_style (NULL, "background");
		if (s) {
			if (app->parse_color (s, &color))
				bg = color;
		}
	}
	if (cpp_class) {
		const char *s = app->fetch_style ((char*) cpp_class, "background");
		if (s) {
			if (app->parse_color (s, &color))
				bg = color;
		}
	}
	if (style_class) {
		snprintf (tmp, 255, "#%s", style_class);
		const char *s = app->fetch_style (tmp, "background");
		if (s) {
			if (app->parse_color (s, &color))
				bg = color;
		}
	}
	if (name) {
		snprintf (tmp, 255, ".%s", name);
		const char *s = app->fetch_style (tmp, "background");
		if (s) {
			if (app->parse_color (s, &color))
				bg = color;
		}
	}

	//--------------------
	// Foreground color.
	//
	if (true) {
		const char *s = app->fetch_style (NULL, "foreground");
		if (s) {
			if (app->parse_color (s, &color))
				fg = color;
		}
	}
	if (cpp_class) {
		const char *s = app->fetch_style ((char*) cpp_class, "foreground");
		if (s) {
			if (app->parse_color (s, &color))
				fg = color;
		}
	}
	if (style_class) {
		snprintf (tmp, 255, "#%s", style_class);
		const char *s = app->fetch_style (tmp, "foreground");
		if (s) {
			if (app->parse_color (s, &color))
				fg = color;
		}
	}
	if (name) {
		snprintf (tmp, 255, ".%s", name);
		const char *s = app->fetch_style (tmp, "foreground");
		if (s) {
			if (app->parse_color (s, &color))
				fg = color;
		}
	}

	//--------------------
	// Border color.
	//
	if (true) {
		const char *s = app->fetch_style (NULL, "border-color");
		if (s) {
			if (app->parse_color (s, &color))
				border_color = color;
		}
	}
	if (cpp_class) {
		const char *s = app->fetch_style ((char*) cpp_class, "border-color");
		if (s) {
			if (app->parse_color (s, &color))
				border_color = color;
		}
	}
	if (style_class) {
		snprintf (tmp, 255, "#%s", style_class);
		const char *s = app->fetch_style (tmp, "border-color");
		if (s) {
			if (app->parse_color (s, &color))
				border_color = color;
		}
	}
	if (name) {
		snprintf (tmp, 255, ".%s", name);
		const char *s = app->fetch_style (tmp, "border-color");
		if (s) {
			if (app->parse_color (s, &color))
				border_color = color;
		}
	}

	//--------------------
	// Border width.
	//
	if (true) {
		const char *s = app->fetch_style (NULL, "border-width");
		if (s) 
			border = atoi (s);
	}
	if (cpp_class) {
		const char *s = app->fetch_style ((char*) cpp_class, "border-width");
		if (s) 
			border = atoi (s);
	}
	if (style_class) {
		snprintf (tmp, 255, "#%s", style_class);
		const char *s = app->fetch_style (tmp, "border-width");
		if (s) 
			border = atoi (s);
	}
	if (name) {
		snprintf (tmp, 255, ".%s", name);
		const char *s = app->fetch_style (tmp, "border-width");
		if (s) 
			border = atoi (s);
	}
}

void 
Widget::redraw (void) 
{ 
	if (!window)
		return;

	Rect r = get_bounds (); 

#if 1
	XExposeEvent event;
	event.type = Expose;
	event.send_event = true;
	event.display = app->dpy;
	event.window = window;
	event.x = 0;
	event.y = 0;
	event.width = rect.size.width;
	event.height = rect.size.height;
	event.count = 0;

	XSendEvent (app->dpy, window, true, ExposureMask, (XEvent*) &event);
#else
	draw (r);
#endif
}

