/*=========================================================================
 * This file is part of TouchWidgets, which is a widget library that
 * supports touch and animation.
 *
 * Copyright (C) 2004,2008-2010,2013 Zachary Smith, veritas@comcast.net.
 *
 * TouchWidgets is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TouchWidgets is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with TouchWidgets.  If not, see <http://www.gnu.org/licenses/>.
 *=======================================================================*/

#ifndef _TOUCHWIDGETS_WIDGET_H
#define _TOUCHWIDGETS_WIDGET_H

#include <TouchWidgets/Common.h>
#include <TouchWidgets/Object.h>
#include <TouchWidgets/Controller.h>

namespace TouchWidgets {

#define TOUCHWIDGET_MIN_DIMENSION (37)

class Widget : public Object
{
protected:
	Window window;

	Rect rect;

	//--------------------
	// Animation data.
	//
	bool animating;
	Rect animation_origin;
	Rect animation_destination;
	unsigned animation_duration;
	unsigned animation_counter;
	Object *animation_delegate; // Whom to tell that animation ended.

	Widget *parent;
	Widget **children;
	int max_children;
	Widget *next; 
        void reallocate_children (int);

	void draw_gradient_background (Rect clip);
	void draw_gradient_in_rect (Rect clip, Rect r);
	void draw_3d_border (Rect clip, bool outward);

	bool hidden;

public:
	Widget (void);
	~Widget (void);

	bool automatic_styling;
	int n_children;
	RGB bg, fg, bg_gradient, highlight_color, selected_color, border_color;
	int border;

	void redraw (void);

        virtual const char* class_name () { return "Widget"; }

	Widget *get_parent (void) { return parent; }

	void clip_into (Rect r, GC gc);

	//--------------------
	// Animation data.
	//
	void animate_to_rect (Rect r, float duration);
	void print_animation_diagnostics (void) {
		printf ("Widget %s (%s) animation: %d of %d.\n", name, class_name(), animation_counter, animation_duration);
	}
	void set_animation_delegate (Object *d) { if (d) animation_delegate = d; }
	void animation_completed (void);
	void generate_next_animation_rect (void);
	bool is_animating (void) { return animating; }

	void set_backing_store (bool on);

	void clear (void);
	bool inside (int,int);
	bool inside (Point);

	Rect get_rect (void);
	Rect get_bounds (void);

	void update_styles (void);

        // Virtual methods
	virtual Window get_window (void);

	void create_window (Window parent_window);
	virtual void destroy_window ();

	void create_windows_for_self_and_children (Window parent_window);

	void move (Point p);
	void set_rect (Rect r);

	void set_hidden (bool);
	bool get_hidden (void) { return hidden; }

	virtual void draw (Rect r);

        virtual void process_event (XEvent *) { }

        virtual void touches_start (Point*, int n_points) { }
        virtual void touches_continue (Point*, int n_points) { }
        virtual void touches_end (Point*, int n_points) { }
        
	void add_child (Widget*);
	void remove_child (Widget*);
	void insert_child_at (Widget*, int);
	Widget *get_nth_child (int);

	void remove_from_parent (void);

	static Point convert_absolute_point_to_relative (Point, Widget*);

        virtual void perform_internal_animation (void) { }

        virtual void message (Message *m); // Let subclasses define this.

	Widget* identify_child_at_relative_point (Point point, Point *return_relative_point);
	Widget* identify_child_for_absolute_point (Point point, Point *return_relative_point);

};

}

#endif

